# -*- coding: utf-8 -*-

""" Some more experimental video calibration routines. Use at your own risk. """

# pylint: disable=too-many-locals

import logging
import copy
import numpy as np
import cv2
from scipy.optimize import minimize
import sksurgerycore.transforms.matrix as skcm
import sksurgerycalibration.video.video_calibration_utils as vu
import sksurgerycalibration.video.video_calibration_cost_functions as vcf
import sksurgerycalibration.video.video_calibration_metrics as vm
import sksurgerycalibration.video.video_calibration_wrapper as vc

LOGGER = logging.getLogger(__name__)


def mono_video_calibration_expt(ids,
                                object_points,
                                image_points,
                                image_size):
    """
    Experimental.

    :param ids: Vector of ndarrays containing integer point ids.
    :param object_points: Vector of Vectors of 1x3 object points, float32
    :param image_points:  Vector of Vectors of 1x2 object points, float32
    :param image_size: (x, y) tuple, size in pixels, e.g. (1920, 1080)
    :return: rms, camera_matrix, dist_coeffs, rvecs, tvecs
    """
    # First do a standard OpenCV calibration, using the wrapper in this project.
    _, camera_matrix, dist_coeffs, rvecs, tvecs = \
        vc.mono_video_calibration(object_points,
                                  image_points,
                                  image_size)

    number_of_views = len(rvecs)
    number_of_intrinsic_parameters = 4 + dist_coeffs.shape[1]
    number_of_extrinsic_parameters = 6 * len(rvecs)

    # Now alternately optimise intrinsics and extrinsics
    for _ in range(0, 5):

        # First optimise intrinsics, via projection error.
        x_0 = np.zeros(number_of_intrinsic_parameters)
        x_0[0] = camera_matrix[0][0]
        x_0[1] = camera_matrix[1][1]
        x_0[2] = camera_matrix[0][2]
        x_0[3] = camera_matrix[1][2]
        for i in range(4, number_of_intrinsic_parameters):
            x_0[i] = dist_coeffs[0][i - 4]

        res = minimize(vcf.mono_reproj_err_for_intrin, x_0,
                       args=(object_points,
                             image_points,
                             rvecs,
                             tvecs),
                       method='Nelder-Mead',
                       tol=1e-3,
                       options={'disp': False, 'maxiter': 1000})
        x_1 = res.x
        camera_matrix[0][0] = x_1[0]
        camera_matrix[1][1] = x_1[1]
        camera_matrix[0][2] = x_1[2]
        camera_matrix[1][2] = x_1[3]
        for i in range(4, number_of_intrinsic_parameters):
            dist_coeffs[0][i - 4] = x_1[i]

        # Now optimise extrinsics, via triangulation.
        x_0 = np.zeros(number_of_extrinsic_parameters)
        for i in range(0, number_of_views):
            x_0[6 * i + 0] = rvecs[i][0]
            x_0[6 * i + 1] = rvecs[i][1]
            x_0[6 * i + 2] = rvecs[i][2]
            x_0[6 * i + 3] = tvecs[i][0]
            x_0[6 * i + 4] = tvecs[i][1]
            x_0[6 * i + 5] = tvecs[i][2]

        res = minimize(vcf.mono_recon_err_for_ext, x_0,
                       args=(
                           ids,
                           object_points,
                           image_points,
                           camera_matrix,
                           dist_coeffs
                       ),
                       method='Nelder-Mead',
                       tol=1e-3,
                       options={'disp': False, 'maxiter': 1000})
        x_1 = res.x
        for i in range(0, number_of_views):
            rvecs[i][0] = x_1[6 * i + 0]
            rvecs[i][1] = x_1[6 * i + 1]
            rvecs[i][2] = x_1[6 * i + 2]
            tvecs[i][0] = x_1[6 * i + 3]
            tvecs[i][1] = x_1[6 * i + 4]
            tvecs[i][2] = x_1[6 * i + 5]

    sse, num = vm.compute_mono_2d_err(object_points,
                                      image_points,
                                      rvecs,
                                      tvecs,
                                      camera_matrix,
                                      dist_coeffs)

    mse = sse / num
    final_rms = np.sqrt(mse)

    return final_rms, camera_matrix, dist_coeffs, rvecs, tvecs


def stereo_video_calibration_expt(left_ids,
                                  left_object_points,
                                  left_image_points,
                                  right_ids,
                                  right_object_points,
                                  right_image_points,
                                  image_size):
    """
    Experimental.

    :param left_ids: Vector of ndarrays containing integer point ids.
    :param left_object_points: Vector of Vectors of 1x3 object points, float32
    :param left_image_points:  Vector of Vectors of 1x2 object points, float32
    :param right_ids: Vector of ndarrays containing integer point ids.
    :param right_object_points: Vector of Vectors of 1x3 object points, float32
    :param right_image_points: Vector of Vectors of 1x2 object points, float32
    :param image_size: (x, y) tuple, size in pixels, e.g. (1920, 1080)
    :return:
    """
    # First do standard OpenCV calibration, using the wrapper in this project.
    s_reproj, s_recon, \
        l_c, l_d, l_rvecs, l_tvecs, \
        r_c, r_d, r_rvecs, r_tvecs, \
        l2r_r, l2r_t, \
        essential, fundamental = \
        vc.stereo_video_calibration(left_ids,
                                    left_object_points,
                                    left_image_points,
                                    right_ids,
                                    right_object_points,
                                    right_image_points,
                                    image_size)

    # For stereo reconstruction error we need common points.
    _, common_object_points, common_left_image_points, \
        common_right_image_points \
        = vu.filter_common_points_all_images(left_ids,
                                             left_object_points,
                                             left_image_points,
                                             right_ids,
                                             right_image_points, 10)

    l2r_rvec = (cv2.Rodrigues(l2r_r.T))[0]

    # Now compute a set of left extrinsics, where the parameters
    # are all relative to the first camera.
    camera_rvecs = copy.deepcopy(l_rvecs)
    camera_tvecs = copy.deepcopy(l_tvecs)
    number_of_frames = len(left_object_points)
    first_world_to_camera = \
        vu.extrinsic_vecs_to_matrix(camera_rvecs[0], camera_tvecs[0])
    first_camera_to_world = np.linalg.inv(first_world_to_camera)
    for i in range(1, number_of_frames):
        extrinsic = \
            vu.extrinsic_vecs_to_matrix(camera_rvecs[i], camera_tvecs[i])
        relative_to_first = np.matmul(extrinsic, first_camera_to_world)
        camera_rvecs[i], camera_tvecs[i] = \
            vu.extrinsic_matrix_to_vecs(relative_to_first)

    # Now we have to create a flat vector of parameters to optimise.
    # Optimsing l2r_r (3), l2r_t(3), l_c (4), r_c (4) + 6 DOF per camera.
    number_of_parameters = 3 + 3 + 4 + 4 + (number_of_frames * 6)
    x_0 = np.zeros(number_of_parameters)
    x_0[0] = l2r_rvec[0][0]
    x_0[1] = l2r_rvec[1][0]
    x_0[2] = l2r_rvec[2][0]
    x_0[3] = l2r_t[0][0]
    x_0[4] = l2r_t[1][0]
    x_0[5] = l2r_t[2][0]
    x_0[6] = l_c[0][0]
    x_0[7] = l_c[1][1]
    x_0[8] = l_c[0][2]
    x_0[9] = l_c[1][2]
    x_0[10] = r_c[0][0]
    x_0[11] = r_c[1][1]
    x_0[12] = r_c[0][2]
    x_0[13] = r_c[1][2]

    for i in range(0, number_of_frames):
        x_0[14 + i * 6 + 0] = camera_rvecs[i][0]
        x_0[14 + i * 6 + 1] = camera_rvecs[i][1]
        x_0[14 + i * 6 + 2] = camera_rvecs[i][2]
        x_0[14 + i * 6 + 3] = camera_tvecs[i][0]
        x_0[14 + i * 6 + 4] = camera_tvecs[i][1]
        x_0[14 + i * 6 + 5] = camera_tvecs[i][2]

    res = minimize(vcf.stereo_2d_and_3d_error, x_0,
                   args=(common_object_points,
                         common_left_image_points,
                         l_d,
                         common_object_points,
                         common_right_image_points,
                         r_d),
                   method='Powell',
                   tol=1e-4,
                   options={'disp': False, 'maxiter': 100000})

    LOGGER.info("Stereo Re-Calibration: success=%s", str(res.success))
    LOGGER.info("Stereo Re-Calibration: msg=%s", str(res.message))

    # Now need to unpack the results, into the same set of vectors,
    # as the stereo_video_calibration, so they are drop-in replacements.
    x_1 = res.x
    l2r_rvec[0][0] = x_1[0]
    l2r_rvec[1][0] = x_1[1]
    l2r_rvec[2][0] = x_1[2]
    l2r_r = (cv2.Rodrigues(l2r_rvec))[0]
    l2r_t[0][0] = x_1[3]
    l2r_t[1][0] = x_1[4]
    l2r_t[2][0] = x_1[5]
    l_c[0][0] = x_1[6]
    l_c[1][1] = x_1[7]
    l_c[0][2] = x_1[8]
    l_c[1][2] = x_1[9]
    r_c[0][0] = x_1[10]
    r_c[1][1] = x_1[11]
    r_c[0][2] = x_1[12]
    r_c[1][2] = x_1[13]

    for i in range(0, number_of_frames):
        camera_rvecs[i][0] = x_1[14 + i * 6 + 0]
        camera_rvecs[i][1] = x_1[14 + i * 6 + 1]
        camera_rvecs[i][2] = x_1[14 + i * 6 + 2]
        camera_tvecs[i][0] = x_1[14 + i * 6 + 3]
        camera_tvecs[i][1] = x_1[14 + i * 6 + 4]
        camera_tvecs[i][2] = x_1[14 + i * 6 + 5]

    # Still need to convert these parameters (which are relative to first
    # camera), into consistent world-to-camera rvecs and tvecs for left/right.
    first_world_to_camera = \
        vu.extrinsic_vecs_to_matrix(camera_rvecs[0], camera_tvecs[0])
    left_to_right = skcm.construct_rigid_transformation(l2r_r, l2r_t)
    for i in range(0, number_of_frames):
        if i == 0:
            left_world_to_camera = first_world_to_camera
        else:
            extrinsic = \
                vu.extrinsic_vecs_to_matrix(camera_rvecs[i], camera_tvecs[i])
            left_world_to_camera = np.matmul(extrinsic, first_world_to_camera)
        l_rvecs[i], l_tvecs[i] = \
            vu.extrinsic_matrix_to_vecs(left_world_to_camera)
        right_world_to_camera = \
            np.matmul(left_to_right, left_world_to_camera)
        r_rvecs[i], r_tvecs[i] = \
            vu.extrinsic_matrix_to_vecs(right_world_to_camera)

    # And recompute stereo RMS re-projection error.
    s_reproj_2 = \
        vm.compute_stereo_2d_err(l2r_r,
                                 l2r_t,
                                 common_object_points,
                                 common_left_image_points,
                                 l_c,
                                 l_d,
                                 common_object_points,
                                 common_right_image_points,
                                 r_c,
                                 r_d,
                                 l_rvecs,
                                 l_tvecs
                                 )

    s_recon_2 = \
        vm.compute_stereo_3d_error(l2r_r,
                                   l2r_t,
                                   common_object_points,
                                   common_left_image_points,
                                   l_c,
                                   l_d,
                                   common_right_image_points,
                                   r_c,
                                   r_d,
                                   l_rvecs,
                                   l_tvecs
                                   )

    LOGGER.info(
        "Stereo Re-Calib: reproj_1=%s, recon_1=%s, reproj_2=%s, recon_2=%s",
        str(s_reproj), str(s_recon), str(s_reproj_2), str(s_recon_2))

    return s_reproj_2, s_recon_2, \
        l_c, l_d, l_rvecs, l_tvecs, \
        r_c, r_d, r_rvecs, r_tvecs, \
        l2r_r, l2r_t, \
        essential, fundamental



# coding=utf-8

""" Functions to run pivot calibration. """

from glob import glob
import numpy as np
from sksurgerycore.configuration.configuration_manager import \
        ConfigurationManager
from sksurgerycalibration.algorithms import pivot as p


def run_pivot_calibration(input_directory, config_file):
    """
    Performs Pivot Calibration using matrices stored in
    separate file input_dir
    """
    configuration = None
    if config_file is not None:
        configurer = ConfigurationManager(config_file)
        configuration = configurer.get_copy()

    input_files = input_directory + '/*'
    file_names = glob(input_files)
    arrays = [np.loadtxt(f) for f in file_names]

    matrices = np.concatenate(arrays)

    number_of_4x4_matrices = int(matrices.size / 16)

    matrices_4x4 = matrices.reshape((number_of_4x4_matrices, 4, 4))

    pointer_offset, pivot_location, residual_error = \
            p.pivot_calibration(matrices_4x4, configuration)

    print("Pointer Offset = ", pointer_offset.reshape((1, 3)))
    print("Pivot Location = ", pivot_location.reshape((1, 3)))
    print("Residual Error = ", residual_error)



# -*- coding: utf-8 -*-

""" Containers for video calibration data. """

import os
import copy
import cv2
import numpy as np
import sksurgerycalibration.video.video_calibration_io as sksio
import sksurgerycalibration.video.video_calibration_hand_eye as heye


class BaseVideoCalibrationData:
    """
    Constructor, no member variables, so just a pure virtual interface.

    Not really necessary if you rely on duck-typing, but at least
    it shows the intention of what derived classes should implement,
    and means we can use this base class to type check against.
    """
    def __init__(self):
        return

    def reinit(self):
        """ Used to clear, re-initialise all member variables. """
        raise NotImplementedError("Derived classes should implement this.")

    def pop(self):
        """ Remove the last view of data. """
        raise NotImplementedError("Derived classes should implement this.")

    def get_number_of_views(self):
        """ Returns the number of views of data. """
        raise NotImplementedError("Derived classes should implement this.")

    def save_data(self, dir_name: str, file_prefix: str):
        """ Writes all contained data to disk. """
        raise NotImplementedError("Derived classes should implement this.")

    def load_data(self, dir_name: str, file_prefix: str):
        """ Loads all contained data from disk. """
        raise NotImplementedError("Derived classes should implement this.")


class TrackingData(BaseVideoCalibrationData):
    """
    Class for storing tracking data.
    """
    def __init__(self):
        super(TrackingData, self).__init__()
        self.device_tracking_array = None
        self.calibration_tracking_array = None
        self.use_quaternions = False
        self.quat_model2hand_array = None
        self.trans_model2hand_array = None
        self.reinit()

    def reinit(self):
        """
        Deletes all data.
        """
        self.device_tracking_array = []
        self.calibration_tracking_array = []

    def push(self, device_tracking, calibration_tracking):
        """
        Stores a pair of tracking data.

        :param device_tracking: transformation for the thing you're tracking
        :param calibration_tracking: transformation for tracked calibration obj
        """
        self.device_tracking_array.append(
            copy.deepcopy(device_tracking))
        self.calibration_tracking_array.append(
            copy.deepcopy(calibration_tracking))

    def pop(self):
        """
        Removes the last (most recent) view of data.
        """
        if self.device_tracking_array:
            self.device_tracking_array.pop(-1)
            self.calibration_tracking_array.pop(-1)

    def get_number_of_views(self):
        """
        Returns the number of views of data.
        :return: int
        """
        return len(self.device_tracking_array)

    def save_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Saves the tracking data to lots of different files.

        :param dir_name: directory to save to
        :param file_prefix: prefix for all files
        """
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        for i in enumerate(self.device_tracking_array):
            device_tracking_file = \
                sksio.get_device_tracking_file_name(dir_name,
                                                    file_prefix,
                                                    i[0])
            if self.device_tracking_array[i[0]] is not None:
                np.savetxt(device_tracking_file,
                           self.device_tracking_array[i[0]], fmt='%.8f')

        for i in enumerate(self.calibration_tracking_array):
            calibration_tracking_file = \
                sksio.get_calibration_tracking_file_name(dir_name,
                                                         file_prefix,
                                                         i[0])
            if self.calibration_tracking_array[i[0]] is not None:
                np.savetxt(calibration_tracking_file,
                           self.calibration_tracking_array[i[0]], fmt='%.8f')

    def load_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Loads tracking data from files.

        :param dir_name: directory to load from
        :param file_prefix: prefix for all files
        """
        self.reinit()
        files = sksio.get_filenames_by_glob_expr(dir_name,
                                                 file_prefix,
                                                 "device_tracking",
                                                 ".txt")
        for file in files:
            device_data = np.loadtxt(file)
            self.device_tracking_array.append(device_data)

        files = sksio.get_filenames_by_glob_expr(dir_name,
                                                 file_prefix,
                                                 "calib_obj_tracking",
                                                 ".txt")
        for file in files:
            calibration_data = np.loadtxt(file)
            self.calibration_tracking_array.append(calibration_data)

    def set_model2hand_arrays(self, use_quaternions=False):
        """
        TODO: Docstring update
        Set the attributes model-to-hand quaternion and translation arrays
        from tracking data.
        """

        self.quat_model2hand_array, self.trans_model2hand_array = \
            heye.set_model2hand_arrays(self.calibration_tracking_array,
                                       self.device_tracking_array,
                                       use_quaternions)


class MonoVideoData(BaseVideoCalibrationData):
    """
    Stores data extracted from each video view of a mono calibration.
    """
    def __init__(self):
        super(MonoVideoData, self).__init__()
        self.images_array = None
        self.ids_arrays = None
        self.object_points_arrays = None
        self.image_points_arrays = None
        self.reinit()

    def reinit(self):
        """
        Deletes all data.
        """
        self.images_array = []
        self.ids_arrays = []
        self.object_points_arrays = []
        self.image_points_arrays = []

    def pop(self):
        """
        Removes the last (most recent) view of data.
        """
        if len(self.images_array) > 1:
            self.images_array.pop(-1)
            self.ids_arrays.pop(-1)
            self.object_points_arrays.pop(-1)
            self.image_points_arrays.pop(-1)

    def push(self, image, ids, object_points, image_points):
        """
        Stores another view of data. Copies data.
        """
        self.images_array.append(copy.deepcopy(image))
        self.ids_arrays.append(copy.deepcopy(ids))
        self.object_points_arrays.append(copy.deepcopy(object_points))
        self.image_points_arrays.append(copy.deepcopy(image_points))

    def get_number_of_views(self):
        """
        Returns the number of views.
        """
        return len(self.images_array)

    def save_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Saves the calibration data to lots of different files.

        :param dir_name: directory to save to
        :param file_prefix: prefix for all files
        """
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        for i in enumerate(self.images_array):
            image_file = sksio.get_images_file_name(dir_name,
                                                    file_prefix,
                                                    i[0])
            cv2.imwrite(image_file, self.images_array[i[0]])
        for i in enumerate(self.ids_arrays):
            id_file = sksio.get_ids_file_name(dir_name,
                                              file_prefix,
                                              i[0])
            np.savetxt(id_file, self.ids_arrays[i[0]], fmt='%d')
        for i in enumerate(self.object_points_arrays):
            object_points_file = sksio.get_objectpoints_file_name(dir_name,
                                                                  file_prefix,
                                                                  i[0])
            reshaped = np.reshape(self.object_points_arrays[i[0]], (-1, 3))
            np.savetxt(object_points_file, reshaped, fmt='%.8f')
        for i in enumerate(self.image_points_arrays):
            image_points_file = sksio.get_imagepoints_file_name(dir_name,
                                                                file_prefix,
                                                                i[0])
            reshaped = np.reshape(self.image_points_arrays[i[0]], (-1, 2))
            np.savetxt(image_points_file, reshaped, fmt='%.8f')

    def load_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Loads the calibration data.

        :param dir_name: directory to load from
        :param file_prefix: prefix for all files
        """
        self.reinit()

        files = sksio.get_filenames_by_glob_expr(dir_name,
                                                 file_prefix,
                                                 "images",
                                                 ".png")
        for file in files:
            image = cv2.imread(file)
            self.images_array.append(image)

        files = sksio.get_filenames_by_glob_expr(dir_name,
                                                 file_prefix,
                                                 "ids",
                                                 ".txt")
        for file in files:
            ids = np.loadtxt(file)
            self.ids_arrays.append(ids)

        files = sksio.get_filenames_by_glob_expr(dir_name,
                                                 file_prefix,
                                                 "object_points",
                                                 ".txt")

        for file in files:
            object_points = np.loadtxt(file)
            reshaped = np.reshape(object_points, (object_points.shape[0], 1, 3))
            self.object_points_arrays.append(reshaped.astype(np.float32))

        files = sksio.get_filenames_by_glob_expr(dir_name,
                                                 file_prefix,
                                                 "image_points",
                                                 ".txt")

        for file in files:
            image_points = np.loadtxt(file)
            reshaped = np.reshape(image_points, (image_points.shape[0], 1, 2))
            self.image_points_arrays.append(reshaped.astype(np.float32))


class StereoVideoData(BaseVideoCalibrationData):
    """
    Stores data extracted from each view of a stereo calibration.
    """
    def __init__(self):
        super(StereoVideoData, self).__init__()
        self.left_data = MonoVideoData()
        self.right_data = MonoVideoData()
        self.reinit()

    def reinit(self):
        """
        Deletes all data.
        """
        self.left_data.reinit()
        self.right_data.reinit()

    def pop(self):
        """
        Removes the last (most recent) view of data.
        """
        self.left_data.pop()
        self.right_data.pop()

    def push(self,
             left_image, left_ids, left_object_points, left_image_points,
             right_image, right_ids, right_object_points, right_image_points):
        """
        Stores another view of data. Copies data.
        """
        self.left_data.push(
            left_image, left_ids, left_object_points, left_image_points)
        self.right_data.push(
            right_image, right_ids, right_object_points, right_image_points)

    def get_number_of_views(self):
        """
        Returns the number of views.
        """
        num_left = self.left_data.get_number_of_views()
        num_right = self.right_data.get_number_of_views()
        if num_left != num_right:
            raise ValueError("Different number of views in left and right??")
        return num_left

    def save_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Saves the calibration data to lots of different files.

        :param dir_name: directory to save to
        :param file_prefix: prefix for all files
        """
        left_prefix = sksio.get_left_prefix(file_prefix)
        self.left_data.save_data(dir_name, left_prefix)
        right_prefix = sksio.get_right_prefix(file_prefix)
        self.right_data.save_data(dir_name, right_prefix)

    def load_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Loads the calibration data.

        :param dir_name: directory to load from
        :param file_prefix: prefix for all files
        """
        self.reinit()
        left_prefix = sksio.get_left_prefix(file_prefix)
        self.left_data.load_data(dir_name, left_prefix)
        right_prefix = sksio.get_right_prefix(file_prefix)
        self.right_data.load_data(dir_name, right_prefix)



# -*- coding: utf-8 -*-

""" Class to do stateful video calibration of a mono camera. """

import copy
import logging
import numpy as np
import sksurgeryimage.calibration.point_detector as pd
import sksurgerycalibration.video.video_calibration_driver_base as vdb
import sksurgerycalibration.video.video_calibration_data as cd
import sksurgerycalibration.video.video_calibration_params as cp
import sksurgerycalibration.video.video_calibration_metrics as cm
import sksurgerycalibration.video.video_calibration_utils as cu
import sksurgerycalibration.video.video_calibration_wrapper as vc

LOGGER = logging.getLogger(__name__)


class MonoVideoCalibrationDriver(vdb.BaseVideoCalibrationDriver):
    """ Class to do stateful video calibration of a mono camera. """
    def __init__(self,
                 point_detector: pd.PointDetector,
                 minimum_points_per_frame: int
                 ):
        """
        Stateful class for mono video calibration.

        :param point_detector: Class derived from PointDetector
        :param minimum_points_per_frame: Minimum number to accept frame
        """
        super(MonoVideoCalibrationDriver, self).\
            __init__(point_detector,
                     minimum_points_per_frame)

        # Create data holders, and parameter holders, specific to Mono.
        calibration_data = cd.MonoVideoData()
        calibration_params = cp.MonoCalibrationParams()

        # Pass them to base class, so base class can access them.
        self._init_internal(calibration_data, calibration_params)

    def grab_data(self,
                  image,
                  device_tracking=None,
                  calibration_object_tracking=None):
        """
        Extracts points, by passing it to the PointDetector.

        This will throw various exceptions if the input data is invalid,
        but will return empty arrays if no points were detected.
        So, no points is not an error. Its an expected condition.

        :param image: RGB image.
        :param device_tracking: transformation for the tracked device
        :param calibration_object_tracking: transformation of tracked
        calibration object
        :return: The number of points grabbed.
        """
        number_of_points = 0

        ids, object_points, image_points = \
            self.point_detector.get_points(image)

        if ids is not None and ids.shape[0] >= self.minimum_points_per_frame:

            ids, image_points, object_points = \
                cu.convert_pd_to_opencv(ids,
                                        object_points,
                                        image_points)

            self.video_data.push(image,
                                 ids,
                                 object_points,
                                 image_points)

            self.tracking_data.push(device_tracking,
                                    calibration_object_tracking)

            number_of_points = image_points.shape[0]

        LOGGER.info("Grabbed: Returning %s points.", str(number_of_points))
        return number_of_points

    def calibrate(self, flags=0):
        """
        Do the video calibration.

        This returns RMS projection error, which is a common metric, but also,
        the reconstruction error. If we have N views, we can take successive
        pairs of views, triangulate points, and see how well they match the
        model. Ideally, both metrics should be small.

        :param flags: OpenCV flags, eg. cv2.CALIB_FIX_ASPECT_RATIO
        :return: RMS projection, reconstruction error.
        """
        proj_err, camera_matrix, dist_coeffs, rvecs, tvecs = \
            vc.mono_video_calibration(
                self.video_data.object_points_arrays,
                self.video_data.image_points_arrays,
                (self.video_data.images_array[0].shape[1],
                 self.video_data.images_array[0].shape[0]),
                flags
            )

        sse, num_samples = \
            cm.compute_mono_3d_err(
                self.video_data.ids_arrays,
                self.video_data.object_points_arrays,
                self.video_data.image_points_arrays,
                rvecs,
                tvecs,
                camera_matrix,
                dist_coeffs
                )
        recon_err = np.sqrt(sse / num_samples)

        self.calibration_params.set_data(camera_matrix,
                                         dist_coeffs,
                                         rvecs,
                                         tvecs)

        LOGGER.info("Calibrated: proj_err=%s, recon_err=%s.",
                    str(proj_err), str(recon_err))
        return proj_err, recon_err, copy.deepcopy(self.calibration_params)

    def iterative_calibration(self,
                              number_of_iterations: int,
                              reference_ids,
                              reference_image_points,
                              reference_image_size,
                              flags: int = 0):
        """
        Does iterative calibration, like Datta 2009.
        """
        proj_err, recon_err, param_copy = self.calibrate(flags=flags)
        cached_images = copy.deepcopy(self.video_data.images_array)

        for i in range(0, number_of_iterations):
            images = copy.deepcopy(cached_images)
            cu.detect_points_in_canonical_space(
                self.point_detector,
                self.minimum_points_per_frame,
                self.video_data,
                images,
                self.calibration_params.camera_matrix,
                self.calibration_params.dist_coeffs,
                reference_ids,
                reference_image_points,
                reference_image_size)
            proj_err, recon_err, param_copy = self.calibrate(flags=flags)
            LOGGER.info("Iterative calibration: %s: proj_err=%s, recon_err=%s.",
                        str(i), str(proj_err), str(recon_err))

        return proj_err, recon_err, param_copy

    def handeye_calibration(self):
        """Do handeye calibration.

        This returns RMS projection error, which is a common metric, but also,
        the reconstruction error. If we have N views, we can take successive
        pairs of views, triangulate points, and see how well they match the
        model. Ideally, both metrics should be small.

        :return: reprojection, reconstruction error
        :rtype: float, float
        """
        self.tracking_data.set_model2hand_arrays()

        proj_err, recon_err, handeye, pattern2marker = \
            vc.mono_handeye_calibration(
                self.video_data.object_points_arrays,
                self.video_data.image_points_arrays,
                self.video_data.ids_arrays,
                self.calibration_params.camera_matrix,
                self.calibration_params.dist_coeffs,
                self.tracking_data.device_tracking_array,
                self.tracking_data.calibration_tracking_array,
                self.calibration_params.rvecs,
                self.calibration_params.tvecs,
                self.tracking_data.quat_model2hand_array,
                self.tracking_data.trans_model2hand_array
            )

        self.calibration_params.set_handeye(handeye, pattern2marker)

        return proj_err, recon_err



# -*- coding: utf-8 -*-

# pylint: disable=unused-import, superfluous-parens, line-too-long, missing-module-docstring, unused-variable, missing-function-docstring, invalid-name

import glob
import pytest
import numpy as np
import cv2
import sksurgeryimage.calibration.chessboard_point_detector as pd
import sksurgerycalibration.video.video_calibration_driver_mono as mc
import sksurgerycalibration.video.video_calibration_driver_stereo as sc
import sksurgerycalibration.video.video_calibration_utils as vu


def get_iterative_reference_data():
    number_of_points = 140
    x_size = 14
    y_size = 10
    pixels_per_square = 20
    reference_ids = np.zeros((number_of_points, 1))
    reference_points = np.zeros((number_of_points, 2))
    counter = 0
    for y_index in range(y_size):
        for x_index in range(x_size):
            reference_ids[counter][0] = counter
            reference_points[counter][0] = (x_index + 2) * pixels_per_square
            reference_points[counter][1] = (y_index + 2) * pixels_per_square
            counter = counter + 1
    reference_image_size = ((x_size + 4) * pixels_per_square, (y_size + 4) * pixels_per_square)
    return reference_ids, reference_points, reference_image_size


def test_chessboard_mono():

    images = []

    files = glob.glob('tests/data/laparoscope_calibration/left/*.png')
    for file in files:
        image = cv2.imread(file)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        images.append(image)

    # This illustrates that the PointDetector sub-class holds the knowledge of the model.
    chessboard_detector = \
        pd.ChessboardPointDetector((14, 10),    # (Width, height), number of internal corners
                                   3,           # Actual square size in mm
                                   (1, 1)       # Scale factors. Here, images are 1920x1080 so no scaling needed.
                                   )            # If images were 1920x540, you'd pass in (1920, 540),
                                                # and PointDetector base class would scale it up.

    # Pass a PointDetector sub-class to the calibration driver.
    calibrator = \
        mc.MonoVideoCalibrationDriver(chessboard_detector, 140)

    # Repeatedly grab data, until you have enough.
    for image in images:
        successful = calibrator.grab_data(image, np.eye(4))
        assert successful > 0
    assert calibrator.is_device_tracked()
    assert not calibrator.is_calibration_target_tracked()

    # Extra checking, as its a unit test
    assert calibrator.get_number_of_views() == 9
    calibrator.pop()
    assert calibrator.get_number_of_views() == 8
    successful = calibrator.grab_data(images[-1])
    assert successful
    assert calibrator.get_number_of_views() == 9

    # Then do calibration
    reproj_err, recon_err, params = calibrator.calibrate()

    # Just for a regression test, checking reprojection error, and recon error.
    # We do appear to get different performance on Linux/Mac
    assert reproj_err < 0.6
    assert recon_err < 0.3

    # Test components of iterative calibration.
    original_image = calibrator.video_data.images_array[0]
    _, _, original_pts = chessboard_detector.get_points(original_image)

    # First ensure undistorting / redistorting points works.
    undistorted_points = cv2.undistortPoints(original_pts,
                                             calibrator.calibration_params.camera_matrix,
                                             calibrator.calibration_params.dist_coeffs,
                                             None,
                                             calibrator.calibration_params.camera_matrix
                                             )
    distorted_pts = vu.distort_points(undistorted_points.reshape((-1, 2)),
                                      calibrator.calibration_params.camera_matrix,
                                      calibrator.calibration_params.dist_coeffs)
    assert np.allclose(original_pts, distorted_pts, rtol=1e-4, atol=1e-4)

    rectify_map_1, rectify_map_2 = cv2.initUndistortRectifyMap(calibrator.calibration_params.camera_matrix,
                                                               calibrator.calibration_params.dist_coeffs,
                                                               None,
                                                               calibrator.calibration_params.camera_matrix,
                                                               (original_image.shape[1], original_image.shape[0]),
                                                               cv2.CV_32FC1
                                                               )
    #undistorted_image = cv2.undistort(original_image, calibrator.calibration_params.camera_matrix, calibrator.calibration_params.dist_coeffs, calibrator.calibration_params.camera_matrix)
    undistorted_image = cv2.remap(original_image, rectify_map_1, rectify_map_2, cv2.INTER_LANCZOS4)

    _, _, undistorted_pts = chessboard_detector.get_points(undistorted_image)
    distorted_pts = vu.distort_points(undistorted_pts, calibrator.calibration_params.camera_matrix, calibrator.calibration_params.dist_coeffs)
    assert np.allclose(original_pts, distorted_pts, rtol=1e-1, atol=1e-1)

    # Test iterative calibration.
    reference_ids, reference_points, reference_image_size = get_iterative_reference_data()

    reproj_err, recon_err, params = calibrator.iterative_calibration(3,
                                                                     reference_ids,
                                                                     reference_points,
                                                                     reference_image_size)
    assert reproj_err < 0.7
    assert recon_err < 0.4


def test_chessboard_stereo():

    left_images = []
    files = glob.glob('tests/data/laparoscope_calibration/left/*.png')
    files.sort()
    for file in files:
        image = cv2.imread(file)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        left_images.append(image)
    assert(len(left_images) == 9)

    right_images = []
    files = glob.glob('tests/data/laparoscope_calibration/right/*.png')
    files.sort()
    for file in files:
        image = cv2.imread(file)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        right_images.append(image)
    assert (len(right_images) == 9)

    chessboard_detector = \
        pd.ChessboardPointDetector((14, 10),
                                   3,
                                   (1, 1)
                                   )

    calibrator = \
        sc.StereoVideoCalibrationDriver(chessboard_detector, 140)

    # Repeatedly grab data, until you have enough.
    for i, _ in enumerate(left_images):
        successful = calibrator.grab_data(left_images[i], right_images[i])
        assert successful > 0
    assert not calibrator.is_device_tracked()
    assert not calibrator.is_calibration_target_tracked()

    # Then do calibration
    reproj_err, recon_err, params = calibrator.calibrate()

    # Just for a regression test, checking reprojection error, and recon error.
    assert reproj_err < 0.7
    assert recon_err < 1.7

    # Test iterative calibration.
    reference_ids, reference_points, reference_image_size = get_iterative_reference_data()

    reproj_err, recon_err, params = calibrator.iterative_calibration(3,
                                                                     reference_ids,
                                                                     reference_points,
                                                                     reference_image_size)
    assert reproj_err < 0.7
    assert recon_err < 1.5



# -*- coding: utf-8 -*-

""" Cost functions for video calibration, used with scipy. """

import numpy as np
import cv2
import sksurgerycalibration.video.video_calibration_utils as vu
import sksurgerycalibration.video.video_calibration_metrics as vm


def stereo_2d_and_3d_error(x_0,
                           left_object_points,
                           left_image_points,
                           left_distortion,
                           right_object_points,
                           right_image_points,
                           right_distortion):
    """
    Private method to compute RMSE cost function, where x_0 contains
    the l2r rvec, l2r tvec, left intrinsics, right intrinsics, and
    left camera extrinsics.
    """
    l2r_rvec = np.zeros((3, 1))
    l2r_rvec[0][0] = x_0[0]
    l2r_rvec[1][0] = x_0[1]
    l2r_rvec[2][0] = x_0[2]
    l2r_rmat = (cv2.Rodrigues(l2r_rvec))[0]
    l2r_tvec = np.zeros((3, 1))
    l2r_tvec[0][0] = x_0[3]
    l2r_tvec[1][0] = x_0[4]
    l2r_tvec[2][0] = x_0[5]
    left_camera = np.eye(3)
    left_camera[0][0] = x_0[6]
    left_camera[1][1] = x_0[7]
    left_camera[0][2] = x_0[8]
    left_camera[1][2] = x_0[9]
    right_camera = np.eye(3)
    right_camera[0][0] = x_0[10]
    right_camera[1][1] = x_0[11]
    right_camera[0][2] = x_0[12]
    right_camera[1][2] = x_0[13]

    sse = 0
    number_of_samples = 0
    number_of_frames = len(left_object_points)

    rvecs = []
    tvecs = []
    for i in range(0, number_of_frames):
        rvec = np.zeros((3, 1))
        rvec[0][0] = x_0[14 + 6 * i + 0]
        rvec[1][0] = x_0[14 + 6 * i + 1]
        rvec[2][0] = x_0[14 + 6 * i + 2]
        tvec = np.zeros((3, 1))
        tvec[0][0] = x_0[14 + 6 * i + 3]
        tvec[1][0] = x_0[14 + 6 * i + 4]
        tvec[2][0] = x_0[14 + 6 * i + 5]
        rvecs.append(rvec)
        tvecs.append(tvec)

    for i in range(0, number_of_frames):

        world_to_left_camera = vu.extrinsic_vecs_to_matrix(rvecs[0], tvecs[0])

        if i > 0:
            offset = vu.extrinsic_vecs_to_matrix(rvecs[i], tvecs[i])
            world_to_left_camera = np.matmul(offset, world_to_left_camera)

        rvec, tvec = vu.extrinsic_matrix_to_vecs(world_to_left_camera)
        rvecs[i] = rvec
        tvecs[i] = tvec

    tmp_sse, tmp_num = vm.compute_stereo_2d_err(l2r_rmat,
                                                l2r_tvec,
                                                left_object_points,
                                                left_image_points,
                                                left_camera,
                                                left_distortion,
                                                right_object_points,
                                                right_image_points,
                                                right_camera,
                                                right_distortion,
                                                rvecs,
                                                tvecs
                                                )
    sse = sse + tmp_sse
    number_of_samples = number_of_samples + tmp_num

    tmp_sse, tmp_num = \
        vm.compute_stereo_3d_error(l2r_rmat,
                                   l2r_tvec,
                                   left_object_points,
                                   left_image_points,
                                   left_camera,
                                   left_distortion,
                                   right_image_points,
                                   right_camera,
                                   right_distortion,
                                   rvecs,
                                   tvecs
                                   )

    sse = sse + tmp_sse
    number_of_samples = number_of_samples + tmp_num
    mse = sse / number_of_samples
    rmse = np.sqrt(mse)
    return rmse


def mono_reproj_err_for_intrin(x_0,
                               object_points,
                               image_points,
                               rvecs,
                               tvecs
                               ):
    """
    Private method to compute SSE projection error over multiple views,
    where x_0 should contain just camera matrix, and then distorion params.
    """
    camera_matrix = np.zeros((3, 3))
    camera_matrix[0][0] = x_0[0]
    camera_matrix[1][1] = x_0[1]
    camera_matrix[0][2] = x_0[2]
    camera_matrix[1][2] = x_0[3]
    number_of_distortion_params = x_0.shape[0] - 4
    distortion_parameters = np.zeros((1, number_of_distortion_params))
    for i in range(0, number_of_distortion_params):
        distortion_parameters[0][i] = x_0[4 + i]

    sse, num_samples = vm.compute_mono_2d_err(object_points,
                                              image_points,
                                              rvecs,
                                              tvecs,
                                              camera_matrix,
                                              distortion_parameters)
    mse = sse / num_samples
    rmse = np.sqrt(mse)
    return rmse


def mono_recon_err_for_ext(x_0,
                           ids,
                           object_points,
                           image_points,
                           camera_matrix,
                           distortion_parameters
                           ):
    """
    Private method to compute SSE reconstruction error over multiple views,
    where x_0 should contain just extrinsic parameters.
    """
    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []
    for i in range(0, number_of_frames):
        rvec = np.zeros((3, 1))
        rvec[0] = x_0[i * 6 + 0]
        rvec[1] = x_0[i * 6 + 1]
        rvec[2] = x_0[i * 6 + 2]
        tvec = np.zeros((3, 1))
        tvec[0] = x_0[i * 6 + 3]
        tvec[1] = x_0[i * 6 + 4]
        tvec[2] = x_0[i * 6 + 5]
        rvecs.append(rvec)
        tvecs.append(tvec)

    #pylint:disable=too-many-function-args
    sse, num_samples = vm.compute_mono_2d_err(ids,
                                              object_points,
                                              image_points,
                                              rvecs,
                                              tvecs,
                                              camera_matrix,
                                              distortion_parameters)
    mse = sse / num_samples
    rmse = np.sqrt(mse)
    return rmse



# -*- coding: utf-8 -*-

# pylint: disable=unused-import, superfluous-parens, line-too-long, missing-module-docstring, unused-variable, missing-function-docstring, invalid-name

import glob
import pytest
import numpy as np
import cv2
import sksurgeryimage.calibration.chessboard_point_detector as pd
import sksurgerycalibration.video.video_calibration_driver_mono as mc
import sksurgerycalibration.video.video_calibration_driver_stereo as sc


def test_chessboard_mono_io():

    images = []

    files = glob.glob('tests/data/laparoscope_calibration/left/*.png')
    for file in files:
        image = cv2.imread(file)
        images.append(image)

    chessboard_detector = pd.ChessboardPointDetector((14, 10), 3, (1, 1))
    calibrator = mc.MonoVideoCalibrationDriver(chessboard_detector, 140)
    for image in images:
        successful = calibrator.grab_data(image, np.eye(4), np.eye(3))
        assert successful > 0

    reproj_err_1, recon_err_1, params_1 = calibrator.calibrate()
    calibrator.save_data('tests/output/test_chessboard_mono_io', '')
    calibrator.save_params('tests/output/test_chessboard_mono_io', '')

    # Now, load data back in, recalibrate, should get same results.
    # Technically, you aren't running the 'grab' bit again.
    # The calibration works off the already extracted points.
    calibrator.load_data('tests/output/test_chessboard_mono_io', '')
    reproj_err_2, recon_err_2, params_2 = calibrator.calibrate()
    assert (np.isclose(reproj_err_1, reproj_err_2))
    assert (np.isclose(recon_err_1, recon_err_2))

    calibrator.load_params('tests/output/test_chessboard_mono_io', '')
    params_3 = calibrator.get_params()
    assert np.allclose(params_2.camera_matrix, params_3.camera_matrix)
    assert np.allclose(params_2.dist_coeffs, params_3.dist_coeffs)
    for i, _ in enumerate(params_3.rvecs):
        assert np.allclose(params_2.rvecs[i], params_3.rvecs[i])
        assert np.allclose(params_2.tvecs[i], params_3.tvecs[i])


def test_chessboard_stereo_io():

    left_images = []
    files = glob.glob('tests/data/laparoscope_calibration/left/*.png')
    files.sort()
    for file in files:
        image = cv2.imread(file)
        left_images.append(image)
    assert(len(left_images) == 9)

    right_images = []
    files = glob.glob('tests/data/laparoscope_calibration/right/*.png')
    files.sort()
    for file in files:
        image = cv2.imread(file)
        right_images.append(image)
    assert (len(right_images) == 9)

    chessboard_detector = \
        pd.ChessboardPointDetector((14, 10),
                                   3,
                                   (1, 1)
                                   )

    calibrator = \
        sc.StereoVideoCalibrationDriver(chessboard_detector, 140)

    for i, _ in enumerate(left_images):
        successful = calibrator.grab_data(left_images[i], right_images[i], np.eye(4), np.eye(3))
        assert successful > 0

    # Then do calibration
    reproj_err_1, recon_err_1, params_1 = calibrator.calibrate()
    calibrator.save_data('tests/output/test_chessboard_stereo_io', '')
    calibrator.save_params('tests/output/test_chessboard_stereo_io', '')

    # Load data, re-do calibration, check for same result.
    calibrator.load_data('tests/output/test_chessboard_stereo_io', '')
    reproj_err_2, recon_err_2, params_2 = calibrator.calibrate()
    assert (np.isclose(reproj_err_1, reproj_err_2))
    assert (np.isclose(recon_err_1, recon_err_2))

    # Now load parameters back in.
    calibrator.load_params('tests/output/test_chessboard_stereo_io', '')
    params_2 = calibrator.get_params()

    assert np.allclose(params_1.l2r_rmat,
                       params_2.l2r_rmat)
    assert np.allclose(params_1.l2r_tvec,
                       params_2.l2r_tvec)
    assert np.allclose(params_1.essential,
                       params_2.essential)
    assert np.allclose(params_1.fundamental,
                       params_2.fundamental)



# coding=utf-8

""" Functions to run video calibration. """

import os
import cv2
from sksurgerycore.configuration.configuration_manager import \
        ConfigurationManager
import sksurgeryimage.calibration.chessboard_point_detector as cpd
import sksurgerycalibration.video.video_calibration_driver_mono as mc

# pylint:disable=too-many-nested-blocks,too-many-branches


def run_video_calibration(config_file, save_dir, prefix):
    """
    Performs Video Calibration using OpenCV
    source and scikit-surgerycalibration.

    :param config_file: mandatory location of config file.
    :param save_dir: optional directory name to dump calibrations to.
    :param prefix: file name prefix when saving
    """
    if config_file is None or len(config_file) == 0:
        raise ValueError("Config file must be provided.")
    if save_dir is not None and prefix is None:
        raise ValueError("If you provide -s/--save, "
                         "you must provide -p/--prefix")

    configurer = ConfigurationManager(config_file)
    configuration = configurer.get_copy()

    # For now just doing chessboards.
    # The underlying framework works for several point detectors,
    # but each would have their own parameters etc.
    source = configuration.get("source", 1)
    corners = configuration.get("corners", [14, 10])
    corners = (corners[0], corners[1])
    size = configuration.get("square size in mm", 3)
    min_num_views = configuration.get("minimum number of views", 5)

    cap = cv2.VideoCapture(int(source))
    if not cap.isOpened():
        raise RuntimeError("Failed to open camera.")

    window_size = configuration.get("window size")
    if window_size is not None:
        cap.set(cv2.CAP_PROP_FRAME_WIDTH, window_size[0])
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, window_size[1])
        print("Video feed set to ("
              + str(window_size[0]) + " x " + str(window_size[1]) + ")")
    else:
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        print("Video feed defaults to ("
              + str(width) + " x " + str(height) + ")")

    detector = cpd.ChessboardPointDetector(corners, size)
    calibrator = mc.MonoVideoCalibrationDriver(detector,
                                               corners[0] * corners[1])

    print("Press 'q' to quit and 'c' to capture an image.")
    print("Minimum number of views to calibrate:" + str(min_num_views))

    while True:
        _, frame = cap.read()
        cv2.imshow("live image", frame)
        key = cv2.waitKey(10)
        if key == ord('q'):
            break
        if key == ord('c'):
            number_points = calibrator.grab_data(frame)
            if number_points > 0:
                img_pts = calibrator.video_data.image_points_arrays[-1]
                img = cv2.drawChessboardCorners(frame, corners,
                                                img_pts,
                                                number_points)
                cv2.imshow("detected points", img)

                number_of_views = calibrator.get_number_of_views()
                print("Number of frames = " + str(number_of_views))

                if number_of_views >= min_num_views:
                    proj_err, recon_err, params = calibrator.calibrate()
                    print("Reprojection (2D) error is:" + str(proj_err))
                    print("Reconstruction (3D) error is:" + str(recon_err))
                    print("Intrinsics are:")
                    print(params.camera_matrix)
                    print("Distortion matrix is:")
                    print(params.dist_coeffs)

                    if save_dir and prefix:

                        if not os.path.isdir(save_dir):
                            os.makedirs(save_dir)

                        calibrator.save_data(save_dir, prefix)
                        calibrator.save_params(save_dir, prefix)
            else:
                print("Failed to detect points")

    cap.release()
    cv2.destroyAllWindows()



# -*- coding: utf-8 -*-

""" Containers for video calibration parameters. """

import os
import copy
import numpy as np
import sksurgerycore.transforms.matrix as sksm
import sksurgerycalibration.video.video_calibration_utils as sksu
import sksurgerycalibration.video.video_calibration_io as sksio


class BaseCalibrationParams:
    """
    Constructor, no member variables, so just a pure virtual interface.

    Not really necessary if you rely on duck-typing, but at least
    it shows the intention of what derived classes should implement,
    and means we can use this base class to type check against.
    """
    def __init__(self):
        return

    def reinit(self):
        """ Used to clear, re-initialise all member variables. """
        raise NotImplementedError("Derived classes should implement this.")

    def save_data(self, dir_name: str, file_prefix: str):
        """ Writes all contained data to disk. """
        raise NotImplementedError("Derived classes should implement this.")

    def load_data(self, dir_name: str, file_prefix: str):
        """ Loads all contained data from disk. """
        raise NotImplementedError("Derived classes should implement this.")


class MonoCalibrationParams(BaseCalibrationParams):
    """
    Holds a set of intrinsic and extrinsic camera parameters for 1 camera.
    """
    def __init__(self):
        super(MonoCalibrationParams, self).__init__()
        self.camera_matrix = None
        self.dist_coeffs = None
        self.rvecs = None
        self.tvecs = None
        self.handeye_matrix = None
        self.pattern2marker_matrix = None
        self.reinit()

    def reinit(self):
        """
        Resets data, to identity/empty arrays etc.
        """
        self.camera_matrix = np.eye(3)
        self.dist_coeffs = np.zeros((1, 5))
        self.rvecs = []
        self.tvecs = []
        self.handeye_matrix = np.eye(4)
        self.pattern2marker_matrix = np.eye(4)

    def set_data(self, camera_matrix, dist_coeffs, rvecs, tvecs):
        """
        Stores the provided parameters, by taking a copy.
        """
        self.camera_matrix = copy.deepcopy(camera_matrix)
        self.dist_coeffs = copy.deepcopy(dist_coeffs)
        self.rvecs = copy.deepcopy(rvecs)
        self.tvecs = copy.deepcopy(tvecs)

    def set_handeye(self, handeye_matrix, pattern2marker_matrix):
        """
        Stores the provided parameters, by taking a copy.
        """
        self.handeye_matrix = copy.deepcopy(handeye_matrix)
        self.pattern2marker_matrix = copy.deepcopy(pattern2marker_matrix)

    def save_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Saves calibration parameters to a directory.

        :param dir_name: directory to save to
        :param file_prefix: prefix for all files
        """
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        intrinsics_file = sksio.get_intrinsics_file_name(dir_name,
                                                         file_prefix)
        np.savetxt(intrinsics_file, self.camera_matrix, fmt='%.8f')

        dist_coeff_file = sksio.get_distortion_file_name(dir_name,
                                                         file_prefix)
        np.savetxt(dist_coeff_file, self.dist_coeffs, fmt='%.8f')
        for i in enumerate(self.rvecs):
            extrinsics_file = sksio.get_extrinsics_file_name(dir_name,
                                                             file_prefix,
                                                             i[0])
            extrinsics = sksu.extrinsic_vecs_to_matrix(self.rvecs[i[0]],
                                                       self.tvecs[i[0]])
            np.savetxt(extrinsics_file, extrinsics, fmt='%.8f')

    def load_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Loads calibration parameters from a directory.

        :param dir_name: directory to load from
        :param file_prefix: prefix for all files
        """
        self.reinit()

        intrinsics_file = sksio.get_intrinsics_file_name(dir_name,
                                                         file_prefix)
        self.camera_matrix = np.loadtxt(intrinsics_file)

        dist_coeff_file = sksio.get_distortion_file_name(dir_name,
                                                         file_prefix)
        self.dist_coeffs = np.loadtxt(dist_coeff_file)

        extrinsic_files = sksio.get_extrinsic_file_names(dir_name,
                                                         file_prefix)
        for file in extrinsic_files:
            extrinsics = np.loadtxt(file)
            rvec, tvec = sksu.extrinsic_matrix_to_vecs(extrinsics)
            self.rvecs.append(rvec)
            self.tvecs.append(tvec)


class StereoCalibrationParams(BaseCalibrationParams):
    """
    Holds a pair of MonoCalibrationParams, and the left-to-right transform.
    """
    def __init__(self):
        super(StereoCalibrationParams, self).__init__()
        self.left_params = None
        self.right_params = None
        self.l2r_rmat = None
        self.l2r_tvec = None
        self.essential = None
        self.fundamental = None
        self.reinit()

    def reinit(self):
        """
        Resets data, to identity/empty arrays etc.
        """
        self.left_params = MonoCalibrationParams()
        self.right_params = MonoCalibrationParams()
        self.l2r_rmat = np.eye(3)
        self.l2r_tvec = np.zeros((3, 1))
        self.essential = np.eye(3)
        self.fundamental = np.eye(3)

    # pylint: disable=too-many-arguments
    def set_data(self,
                 left_cam_matrix, left_dist_coeffs, left_rvecs, left_tvecs,
                 right_cam_matrix, right_dist_coeffs, right_rvecs, right_tvecs,
                 l2r_rmat, l2r_tvec, essential, fundamental
                 ):
        """
        Stores the provided parameters, by taking a copy.
        """
        self.left_params.set_data(left_cam_matrix,
                                  left_dist_coeffs,
                                  left_rvecs,
                                  left_tvecs)
        self.right_params.set_data(right_cam_matrix,
                                   right_dist_coeffs,
                                   right_rvecs,
                                   right_tvecs)
        self.l2r_rmat = copy.deepcopy(l2r_rmat)
        self.l2r_tvec = copy.deepcopy(l2r_tvec)
        self.essential = copy.deepcopy(essential)
        self.fundamental = copy.deepcopy(fundamental)

    def set_handeye(self, left_handeye_matrix, left_pattern2marker_matrix,
                    right_handeye_matrix, right_pattern2marker_matrix):
        """
        Call the left/right set_handeye methods.
        """
        self.left_params.set_handeye(
            left_handeye_matrix, left_pattern2marker_matrix)

        self.right_params.set_handeye(
            right_handeye_matrix, right_pattern2marker_matrix)

    def get_l2r_as_4x4(self):
        """
        Extracts the left-to-right transform as 4x4 matrix.
        """
        return sksm.construct_rigid_transformation(self.l2r_rmat, self.l2r_tvec)

    def save_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Saves calibration parameters to a directory.

        :param dir_name: directory to save to
        :param file_prefix: prefix for all files
        """
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        left_prefix = sksio.get_left_prefix(file_prefix)
        self.left_params.save_data(dir_name, left_prefix)
        right_prefix = sksio.get_right_prefix(file_prefix)
        self.right_params.save_data(dir_name, right_prefix)

        l2r_file = sksio.get_l2r_file_name(dir_name, file_prefix)
        np.savetxt(l2r_file, self.get_l2r_as_4x4(), fmt='%.8f')

        ess_file = sksio.get_essential_matrix_file_name(dir_name,
                                                        file_prefix)
        np.savetxt(ess_file, self.essential, fmt='%.8f')

        fun_file = sksio.get_fundamental_matrix_file_name(dir_name,
                                                          file_prefix)
        np.savetxt(fun_file, self.fundamental, fmt='%.8f')

    def load_data(self,
                  dir_name: str,
                  file_prefix: str
                  ):
        """
        Loads calibration parameters from a directory.

        :param dir_name: directory to load from
        :param file_prefix: prefix for all files
        """
        self.reinit()

        left_prefix = sksio.get_left_prefix(file_prefix)
        self.left_params.load_data(dir_name, left_prefix)
        right_prefix = sksio.get_right_prefix(file_prefix)
        self.right_params.load_data(dir_name, right_prefix)

        l2r_file = sksio.get_l2r_file_name(dir_name, file_prefix)
        stereo_ext = np.loadtxt(l2r_file)

        self.l2r_rmat = stereo_ext[0:3, 0:3]
        tmp = stereo_ext[0:3, 3]
        self.l2r_tvec[0][0] = tmp[0]
        self.l2r_tvec[1][0] = tmp[1]
        self.l2r_tvec[2][0] = tmp[2]

        ess_file = sksio.get_essential_matrix_file_name(dir_name,
                                                        file_prefix)
        self.essential = np.loadtxt(ess_file)

        fun_file = sksio.get_fundamental_matrix_file_name(dir_name,
                                                          file_prefix)
        self.fundamental = np.loadtxt(fun_file)



# -*- coding: utf-8 -*-

""" Base class for our mono and stereo video camera calibration drivers. """

import copy
import logging
import sksurgeryimage.calibration.point_detector as pd
import sksurgerycalibration.video.video_calibration_data as vcd
import sksurgerycalibration.video.video_calibration_params as vcp
import sksurgerycalibration.video.video_calibration_utils as vcu

LOGGER = logging.getLogger(__name__)


class BaseVideoCalibrationDriver:
    """
    Base class for video calibration drivers.
    """
    def __init__(self,
                 point_detector: pd.PointDetector,
                 minimum_points_per_frame: int
                 ):
        """
        Base class for video calibration drivers.

        This class expects calling code to decide how many images are
        required to calibrate, and also, when to call reinit.

        The PointDetector is passed in using Dependency Injection.
        So, the PointDetector can be anything, like chessboards, ArUco,
        CharUco etc.

        This does mean that the underlying code can handle variable numbers
        of points in each view. OpenCV calibration math does this anyway.

        :param point_detector: Class derived from PointDetector
        :param minimum_points_per_frame: Minimum number to accept frame
        """
        self.point_detector = point_detector
        self.minimum_points_per_frame = minimum_points_per_frame
        self.tracking_data = vcd.TrackingData()
        self.video_data = None
        self.calibration_params = None
        LOGGER.info("Constructed: Points per view=%s",
                    str(self.minimum_points_per_frame))

    def _init_internal(self,
                       video_data: vcd.BaseVideoCalibrationData,
                       calibration_params: vcp.BaseCalibrationParams):
        """
        Derived classes must call this, to assign to

        - self.calibration_data
        - self.calibration_params
        """
        self.video_data = video_data
        self.calibration_params = calibration_params

    def reinit(self):
        """
        Resets this object, which means, removes stored calibration data
        and reset the calibration parameters to identity/zero.
        """
        self.tracking_data.reinit()
        self.video_data.reinit()
        self.calibration_params.reinit()
        LOGGER.info("Reset: Now zero frames.")

    def pop(self):
        """
        Removes the last grabbed view of data.
        """
        self.tracking_data.pop()
        self.video_data.pop()
        LOGGER.info("Popped: Now %s views.", str(self.get_number_of_views()))

    def get_number_of_views(self):
        """
        Returns the current number of stored views.

        :return: number of views
        """
        return self.video_data.get_number_of_views()

    def calibrate(self, flags=0):
        """
        Do the video calibration. Derived classes must implement this.
        """
        raise NotImplementedError("Derived classes must implement this.")

    def save_data(self,
                  dir_name: str,
                  file_prefix: str):
        """
        Saves the data to the given dir_name, with file_prefix.
        """
        self.tracking_data.save_data(dir_name, file_prefix)
        self.video_data.save_data(dir_name, file_prefix)

    def load_data(self,
                  dir_name: str,
                  file_prefix: str):
        """
        Loads the data from dir_name, and populates this object.
        """
        self.tracking_data.load_data(dir_name, file_prefix)
        self.video_data.load_data(dir_name, file_prefix)

    def save_params(self,
                    dir_name: str,
                    file_prefix: str):
        """
        Saves the calibration parameters to dir_name, with file_prefix.
        """
        self.calibration_params.save_data(dir_name, file_prefix)

    def load_params(self,
                    dir_name: str,
                    file_prefix: str):
        """
        Loads the calibration params from dir_name, using file_prefix.
        """
        self.calibration_params.load_data(dir_name, file_prefix)

    def get_params(self):
        """
        Copies and returns the parameters.
        """
        return copy.deepcopy(self.calibration_params)

    def get_video_data(self):
        """
        Copies and returns the video data.
        """
        return copy.deepcopy(self.video_data)

    def get_tracking_data(self):
        """
        Copies and returns the tracking data.
        """
        return copy.deepcopy(self.tracking_data)

    def is_device_tracked(self):
        """
        Returns True if we have tracking data for the device.
        """
        result = \
            vcu.array_contains_tracking_data(
                self.tracking_data.device_tracking_array
            )
        return result

    def is_calibration_target_tracked(self):
        """
        Returns True if we have tracking data for the calibration target.
        """
        result = \
            vcu.array_contains_tracking_data(
                self.tracking_data.calibration_tracking_array
            )
        return result



# -*- coding: utf-8 -*-

"""
Tests for calibration_manager.
"""

import glob
import cv2
import pytest
import numpy as np
import sksurgeryimage.calibration.chessboard_point_detector as pd
import sksurgeryimage.calibration.charuco_plus_chessboard_point_detector as chpd
import sksurgerycalibration.video as vidcal


def test_set_model2hand_arrays():
    """Test that model2hand arrays are set without raising an error"""
    chessboard_detector = pd.ChessboardPointDetector((14, 10), 3, (1, 1))

    stereo_calib = \
        vidcal.video_calibration_driver_stereo.StereoVideoCalibrationDriver(
            chessboard_detector, 140)

    tracking_data_dir = 'tests/data/2020_01_20_storz/12_50_30'
    file_prefix = 'calib'

    stereo_calib.load_data(tracking_data_dir, file_prefix)

    stereo_calib.tracking_data.set_model2hand_arrays()


def load_images_from_glob(glob_pattern):
    """ Load images from files based on a glob pattern. """
    images = []
    files = glob.glob(glob_pattern)
    files.sort()
    for file in files:
        image = cv2.imread(file)
        images.append(image)

    return images

def load_tracking_from_glob(glob_pattern):
    """ Load tracking data from files based on a glob pattern. """
    tracking = []
    files = glob.glob(glob_pattern)
    files.sort()
    for file in files:
        data = np.loadtxt(file)
        tracking.append(data)

    return tracking


def test_handeye_calibration_mono():
    """ Load mono data (only using left channel) and tracking, do video and
    handeye calibration, compare results against expected values.
    """
    images = load_images_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.left.*.png')

    device_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.device_tracking.*.txt')

    obj_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.calib_obj_tracking.*.txt')

    assert len(images) == 10

    assert len(device_tracking) == 10
    assert len(obj_tracking) == 10

    min_number_of_points_per_image = 50
    detector = \
        chpd.CharucoPlusChessboardPointDetector(error_if_no_chessboard=False)

    calibrator = \
        vidcal.video_calibration_driver_mono.MonoVideoCalibrationDriver(
            detector, min_number_of_points_per_image)

    # Grab data from images/tracking arrays
    for image, device, calib_obj in zip(images, device_tracking, obj_tracking):
        successful = calibrator.grab_data(image, device, calib_obj)
        assert successful > 0

    reproj_err_1, recon_err_1, _ = calibrator.calibrate()

    one_pixel = 1
    assert reproj_err_1 < one_pixel
    assert recon_err_1 < one_pixel

    proj_err, recon_err = calibrator.handeye_calibration()

    print(f'Reproj err {proj_err}')
    print(f'Recon err {recon_err}')

    # These values are taken from a previous successful run
    # Not objective measures of correctness
    expected_reproj_error = 14.93867
    expected_recon_error = 5.530625

    assert proj_err == pytest.approx(expected_reproj_error, rel=1e-4)
    assert recon_err == pytest.approx(expected_recon_error, rel=1e-4)


def test_handeye_calibration_stereo():
    """ Load Stereo data and tracking, do video and
    handeye calibration, compare results against expected values.
    """
    left_images = load_images_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.left.*.png')

    right_images = load_images_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.right.*.png')

    device_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.device_tracking.*.txt')

    obj_tracking = load_tracking_from_glob(
        'tests/data/2020_01_20_storz/12_50_30/calib.calib_obj_tracking.*.txt')

    assert len(left_images) == 10
    assert len(right_images) == 10
    assert len(device_tracking) == 10
    assert len(obj_tracking) == 10

    min_number_of_points_per_image = 50
    detector = \
        chpd.CharucoPlusChessboardPointDetector(error_if_no_chessboard=False)

    calibrator = \
        vidcal.video_calibration_driver_stereo.StereoVideoCalibrationDriver(
            detector, min_number_of_points_per_image)

    # Grab data from images/tracking arrays
    for left, right, device, calib_obj in \
        zip(left_images, right_images, device_tracking, obj_tracking):
        successful = calibrator.grab_data(left, right, device, calib_obj)
        assert successful > 0

    reproj_err_1, recon_err_1, _ = calibrator.calibrate()

    one_pixel = 1
    assert reproj_err_1 < one_pixel
    assert recon_err_1 < one_pixel

    proj_err, recon_err = calibrator.handeye_calibration()

    print(f'Reproj err {proj_err}')
    print(f'Recon err {recon_err}')

    # These values are taken from a previous successful run
    # Not objective measures of correctness
    expected_reproj_error = 13.452010
    expected_recon_error = 1.394778

    assert proj_err == pytest.approx(expected_reproj_error, rel=1e-4)
    assert recon_err == pytest.approx(expected_recon_error, rel=1e-4)


def test_load_data_stereo_calib():
    """ Load tracking and image data from test directory. """
    chessboard_detector = pd.ChessboardPointDetector((14, 10), 3, (1, 1))

    stereo_calib = \
        vidcal.video_calibration_driver_stereo.StereoVideoCalibrationDriver(
            chessboard_detector, 140)

    tracking_data_dir = 'tests/data/2020_01_20_storz/12_50_30'
    file_prefix = 'calib'

    stereo_calib.load_data(tracking_data_dir, file_prefix)

    assert len(stereo_calib.tracking_data.device_tracking_array) == 10
    assert len(stereo_calib.tracking_data.calibration_tracking_array) == 10
    assert len(stereo_calib.video_data.left_data.images_array) == 10
    assert len(stereo_calib.video_data.right_data.images_array) == 10
    


# -*- coding: utf-8 -*-

""" Various utilities, converters etc., to help video calibration. """

# pylint:disable=invalid-name

import collections
import numpy as np
import cv2
import sksurgerycore.transforms.matrix as skcm


def convert_numpy2d_to_opencv(image_points):
    """
    Converts numpy array to Vector of 1x2 vectors containing float32.

    :param image_points: numpy [Mx2] array.
    :return: vector (length M), of 1x2 vectors of float32.
    """
    return np.reshape(image_points, (-1, 1, 2)).astype(np.float32)


def convert_numpy3d_to_opencv(object_points):
    """
    Converts numpy array to Vector of 1x3 vectors containing float32.

    :param object_points: numpy [Mx3] array.
    :return: vector (length M), of 1x3 vectors of float32.
    """
    return np.reshape(object_points, (-1, 1, 3)).astype(np.float32)


def extrinsic_vecs_to_matrix(rvec, tvec):
    """
    Method to convert rvec and tvec to a 4x4 matrix.

    :param rvec: [3x1] ndarray, Rodrigues rotation params
    :param rvec: [3x1] ndarray, translation params
    :return: [3x3] ndarray, Rotation Matrix
    """
    rotation_matrix = (cv2.Rodrigues(rvec))[0]
    transformation_matrix = \
        skcm.construct_rigid_transformation(rotation_matrix, tvec)
    return transformation_matrix


def extrinsic_matrix_to_vecs(transformation_matrix):
    """
    Method to convert a [4x4] rigid body matrix to an rvec and tvec.

    :param transformation_matrix: [4x4] rigid body matrix.
    :return [3x1] Rodrigues rotation vec, [3x1] translation vec
    """
    rmat = transformation_matrix[0:3, 0:3]
    rvec = (cv2.Rodrigues(rmat))[0]
    tvec = np.ones((3, 1))
    tvec[0:3, 0] = transformation_matrix[0:3, 3]
    return rvec, tvec


def filter_common_points_per_image(left_ids,
                                   left_object_points,
                                   left_image_points,
                                   right_ids,
                                   right_image_points,
                                   minimum_points
                                   ):
    """
    For stereo calibration, we need common points in left and right.
    Remember that a point detector, may provide different numbers of
    points for left and right, and they may not be sorted.

    :param left_ids: ndarray of integer point ids
    :param left_object_points: Vector of Vector of 1x3 float 32
    :param left_image_points: Vector of Vector of 1x2 float 32
    :param right_ids: ndarray of integer point ids
    :param right_image_points: Vector of Vector of 1x2 float 32
    :param minimum_points: the number of minimum common points to accept
    :return: common ids, object_points, left_image_points, right_image_points
    """

    # Filter obvious duplicates first.
    non_duplicate_left = np.asarray(
        [item for item, count in
         collections.Counter(left_ids).items() if count == 1])
    non_duplicate_right = np.asarray(
        [item for item, count in
         collections.Counter(right_ids).items() if count == 1])

    filtered_left = left_ids[
        np.isin(left_ids, non_duplicate_left)]
    filtered_right = right_ids[
        np.isin(right_ids, non_duplicate_right)]

    # Now find common points in left and right.
    ids = np.intersect1d(filtered_left, filtered_right)
    ids = np.sort(ids)

    if len(ids) < minimum_points:
        raise ValueError("Not enough common points in left and right images.")

    common_ids = \
        np.zeros((len(ids), 1), dtype=np.int)
    common_object_points = \
        np.zeros((len(ids), 1, 3), dtype=np.float32)
    common_left_image_points = \
        np.zeros((len(ids), 1, 2), dtype=np.float32)
    common_right_image_points = \
        np.zeros((len(ids), 1, 2), dtype=np.float32)

    counter = 0
    for position in ids:
        left_location = np.where(left_ids == position)
        common_ids[counter] = left_ids[left_location[0][0]]
        common_object_points[counter] \
            = left_object_points[left_location[0][0]]
        common_left_image_points[counter] \
            = left_image_points[left_location[0][0]]
        right_location = np.where(right_ids == position)
        common_right_image_points[counter] \
            = right_image_points[right_location[0][0]]
        counter = counter + 1

    number_of_left = len(common_left_image_points)
    number_of_right = len(common_right_image_points)

    if number_of_left != number_of_right:
        raise ValueError("Unequal number of common points in left and right.")

    return common_ids, common_object_points, common_left_image_points, \
        common_right_image_points


def filter_common_points_all_images(left_ids,
                                    left_object_points,
                                    left_image_points,
                                    right_ids,
                                    right_image_points,
                                    minimum_points
                                    ):
    """
    Loops over each images's data, filtering per image.
    See: filter_common_points_per_image
    :return: Vectors of outputs from filter_common_points_per_image
    """
    common_ids = []
    common_object_points = []
    common_left_image_points = []
    common_right_image_points = []

    # pylint:disable=consider-using-enumerate
    for counter in range(len(left_ids)):
        c_i, c_o, c_l, c_r = \
            filter_common_points_per_image(left_ids[counter],
                                           left_object_points[counter],
                                           left_image_points[counter],
                                           right_ids[counter],
                                           right_image_points[counter],
                                           minimum_points
                                           )
        common_ids.append(c_i)
        common_object_points.append(c_o)
        common_left_image_points.append(c_l)
        common_right_image_points.append(c_r)

    return common_ids, common_object_points, common_left_image_points, \
        common_right_image_points


def convert_pd_to_opencv(ids, object_points, image_points):
    """
    The PointDetectors from scikit-surgeryimage aren't quite compatible
    with OpenCV.
    """
    dims = np.shape(image_points)
    ids = np.reshape(ids, dims[0])
    image_points = np.reshape(image_points, (dims[0], 1, 2))
    image_points = image_points.astype(np.float32)
    object_points = np.reshape(object_points, (-1, 1, 3))
    object_points = object_points.astype(np.float32)
    return ids, image_points, object_points


def array_contains_tracking_data(array_to_check):
    """
    Returns True if the array contains some tracking data.
    """
    result = False
    if array_to_check is not None:
        number_of_items = len(array_to_check)
        if number_of_items > 0:
            found_none = False
            for i in range(0, number_of_items):
                if array_to_check[i] is None:
                    found_none = True
            if not found_none:
                result = True
    return result


def match_points_by_id(ids_1, points_1, ids_2, points_2):
    """
    Returns an ndarray of matched points, matching by their identifier.

    :param ids_1: ndarray [Mx1] list of ids for points_1
    :param points_1: ndarray [Mx2 or 3] of 2D or 3D points
    :param ids_2: ndarray [Nx1] list of ids for points_2
    :param points_2: ndarray [Nx2 or 3] of 2D or 3D points
    :return: ndarray. Number of rows is the number of common points by ids.
    """
    common_ids = np.intersect1d(ids_1, ids_2)
    common_ids = np.sort(common_ids)
    indexes_1 = np.isin(ids_1, common_ids).reshape(-1)
    indexes_2 = np.isin(ids_2, common_ids).reshape(-1)
    points_1_selected = points_1[indexes_1, :]
    points_2_selected = points_2[indexes_2, :]
    result = np.zeros((common_ids.shape[0],
                       points_1_selected.shape[1] +
                       points_2_selected.shape[1]))
    result[:, 0:points_1_selected.shape[1]] \
        = points_1_selected[:, :]
    result[:, points_1_selected.shape[1]:points_1_selected.shape[1] +
           points_2_selected.shape[1]] = points_2_selected[:, :]
    return result


def distort_points(image_points, camera_matrix, distortion_coeffs):
    """
    Distorts image points, reversing the effects of cv2.undistortPoints.

    Slow, but should do for now, for offline calibration at least.

    :param image_points: undistorted image points.
    :param camera_matrix: [3x3] camera matrix
    :param distortion_coeffs: [1x5] distortion coefficients
    :return: distorted points
    """
    distorted_pts = np.zeros(image_points.shape)
    number_of_points = image_points.shape[0]

    # pylint: disable=invalid-name
    for counter in range(number_of_points):
        relative_x = (image_points[counter][0] - camera_matrix[0][2]) \
            / camera_matrix[0][0]
        relative_y = (image_points[counter][1] - camera_matrix[1][2]) \
            / camera_matrix[1][1]
        r2 = relative_x * relative_x + relative_y * relative_y
        radial = (
            1
            + distortion_coeffs[0][0]
            * r2
            + distortion_coeffs[0][1]
            * r2 * r2
            + distortion_coeffs[0][4]
            * r2 * r2 * r2
        )
        distorted_x = relative_x * radial
        distorted_y = relative_y * radial

        distorted_x = distorted_x + (
            2 * distortion_coeffs[0][2]
            * relative_x * relative_y
            + distortion_coeffs[0][3]
            * (r2 + 2 * relative_x * relative_x))

        distorted_y = distorted_y + (
            distortion_coeffs[0][2]
            * (r2 + 2 * relative_y * relative_y)
            + 2 * distortion_coeffs[0][3]
            * relative_x * relative_y)

        distorted_x = distorted_x * camera_matrix[0][0] + camera_matrix[0][2]
        distorted_y = distorted_y * camera_matrix[1][1] + camera_matrix[1][2]

        distorted_pts[counter][0] = distorted_x
        distorted_pts[counter][1] = distorted_y

    return distorted_pts


def map_points_from_canonical_to_original(images_array,
                                          image_index,
                                          video_data,
                                          ids,
                                          object_points,
                                          image_points,
                                          homography,
                                          camera_matrix,
                                          distortion_coeffs):
    """
    Utility method to map image points, detected in a canonical face
    on image, back to the original image space.

    :param images_array:
    :param image_index:
    :param video_data:
    :param ids:
    :param object_points:
    :param image_points:
    :param homography:
    :param camera_matrix:
    :param distortion_coeffs:
    :return:
    """
    inverted_points = \
        cv2.perspectiveTransform(
            image_points.astype(np.float32).reshape(-1, 1, 2),
            np.linalg.inv(homography))
    inverted_points = inverted_points.reshape(-1, 2)

    distorted_pts = distort_points(
        inverted_points,
        camera_matrix,
        distortion_coeffs)

    # Convert back to a format suitable for OpenCV calibration.
    ids, image_points, object_points = \
        convert_pd_to_opencv(ids,
                             object_points,
                             distorted_pts)

    # Store the image, with ids, object_points and image_points.
    video_data.push(images_array[image_index],
                    ids,
                    object_points,
                    image_points)


def detect_points_in_canonical_space(point_detector,
                                     minimum_points_per_frame,
                                     video_data,
                                     images,
                                     camera_matrix,
                                     distortion_coefficients,
                                     reference_ids,
                                     reference_image_points,
                                     reference_image_size
                                     ):
    """
    Method that does the bulk of the heavy lifting in Datta 2009.

    :param point_detector:
    :param minimum_points_per_frame:
    :param video_data:
    :param images:
    :param camera_matrix:
    :param distortion_coefficients:
    :param reference_ids:
    :param reference_image_points:
    :param reference_image_size:
    :return:
    """
    video_data.reinit()
    for j, _ in enumerate(images):

        undistorted = cv2.undistort(
            images[j],
            camera_matrix,
            distortion_coefficients,
            camera_matrix
        )

        ids, obj_pts, img_pts = point_detector.get_points(undistorted)

        if ids is not None and ids.shape[0] >= minimum_points_per_frame:

            common_points = match_points_by_id(ids, img_pts,
                                               reference_ids,
                                               reference_image_points)
            homography, _ = \
                cv2.findHomography(common_points[0:, 0:2],
                                   common_points[0:, 2:4])
            warped = cv2.warpPerspective(undistorted,
                                         homography,
                                         reference_image_size,)

            ids, obj_pts, img_pts = point_detector.get_points(warped)

            if ids is not None and ids.shape[0] >= minimum_points_per_frame:

                map_points_from_canonical_to_original(images,
                                                      j,
                                                      video_data,
                                                      ids,
                                                      obj_pts,
                                                      img_pts,
                                                      homography,
                                                      camera_matrix,
                                                      distortion_coefficients)


# pylint: disable=too-many-arguments
def detect_points_in_stereo_canonical_space(point_detector,
                                            minimum_points_per_frame,
                                            left_video_data,
                                            left_images,
                                            left_camera_matrix,
                                            left_distortion_coeffs,
                                            right_video_data,
                                            right_images,
                                            right_camera_matrix,
                                            right_distortion_coeffs,
                                            reference_ids,
                                            reference_image_points,
                                            reference_image_size
                                            ):
    """
    Method that does the bulk of the heavy lifting in Datta 2009.

    The reason we need a combined stereo one, instead of calling the mono
    one twice, is because at any point, if either left or right channel
    fails feature detection, we need to drop that image from BOTH channels.

    :param point_detector:
    :param minimum_points_per_frame:
    :param left_video_data:
    :param left_images:
    :param left_camera_matrix:
    :param left_distortion_coeffs:
    :param right_video_data:
    :param right_images:
    :param right_camera_matrix:
    :param right_distortion_coeffs:
    :param reference_ids:
    :param reference_image_points:
    :param reference_image_size:
    :return:
    """
    left_video_data.reinit()
    right_video_data.reinit()

    for j, _ in enumerate(left_images):
        left_undistorted = cv2.undistort(
            left_images[j],
            left_camera_matrix,
            left_distortion_coeffs,
            left_camera_matrix
        )
        left_ids, left_obj_pts, left_img_pts = \
            point_detector.get_points(left_undistorted)

        right_undistorted = cv2.undistort(
            right_images[j],
            right_camera_matrix,
            right_distortion_coeffs,
            right_camera_matrix
        )
        right_ids, right_obj_pts, right_img_pts = \
            point_detector.get_points(right_undistorted)

        if left_ids is not None \
                and left_ids.shape[0] >= minimum_points_per_frame \
                and right_ids is not None \
                and right_ids.shape[0] >= minimum_points_per_frame:

            left_common_points = match_points_by_id(left_ids,
                                                    left_img_pts,
                                                    reference_ids,
                                                    reference_image_points)
            left_homography, _ = \
                cv2.findHomography(left_common_points[0:, 0:2],
                                   left_common_points[0:, 2:4])
            left_warped = cv2.warpPerspective(left_undistorted,
                                              left_homography,
                                              reference_image_size,)

            left_ids, left_obj_pts, left_img_pts = \
                point_detector.get_points(left_warped)

            right_common_points = match_points_by_id(right_ids,
                                                     right_img_pts,
                                                     reference_ids,
                                                     reference_image_points)
            right_homography, _ = \
                cv2.findHomography(right_common_points[0:, 0:2],
                                   right_common_points[0:, 2:4])
            right_warped = cv2.warpPerspective(right_undistorted,
                                               right_homography,
                                               reference_image_size,)

            right_ids, right_obj_pts, right_img_pts = \
                point_detector.get_points(right_warped)

            if left_ids is not None \
                    and left_ids.shape[0] >= minimum_points_per_frame \
                    and right_ids is not None \
                    and right_ids.shape[0] >= minimum_points_per_frame:

                map_points_from_canonical_to_original(left_images,
                                                      j,
                                                      left_video_data,
                                                      left_ids,
                                                      left_obj_pts,
                                                      left_img_pts,
                                                      left_homography,
                                                      left_camera_matrix,
                                                      left_distortion_coeffs)

                map_points_from_canonical_to_original(right_images,
                                                      j,
                                                      right_video_data,
                                                      right_ids,
                                                      right_obj_pts,
                                                      right_img_pts,
                                                      right_homography,
                                                      right_camera_matrix,
                                                      right_distortion_coeffs)



""" Functions used by calibration the calibration routines """

from scipy.optimize import least_squares
import numpy

def fit_sphere_least_squares(coordinates, initial_parameters,
                             bounds=
                             ((-numpy.inf, -numpy.inf, -numpy.inf, -numpy.inf),
                              (numpy.inf, numpy.inf, numpy.inf, numpy.inf))):
    """
    Uses scipy's least squares optimisor to fit a sphere to a set
    of 3D Points

    :param coordinates: (x,y,z) n x 3 array of point coordinates
    :param initial parameters: 1 x 4  array containing four initial
        values (centre, and radius)
    :return: x: an array containing the four fitted parameters
    :return: ier: int An integer flag. If it is equal to 1, 2, 3 or 4, the
             solution was found.
    """
    x_values = coordinates[:, 0]
    y_values = coordinates[:, 1]
    z_values = coordinates[:, 2]
    return least_squares(_calculate_residual_sphere, initial_parameters,
                         bounds=bounds,
                         method='trf',
                         jac='3-point',
                         args=(x_values, y_values, z_values))

def _calculate_residual_sphere(parameters, x_values, y_values, z_values):
    """
    Calculates the residual error for an x,y,z coordinates, fitted
    to a sphere with centre and radius defined by the parameters tuple

    :param: A tuple of the parameters to be optimised, should contain
            [x_centre, y_centre, z_centre, radius]
    :param: arrays containing the x,y, and z coordinates.
    :return: The residual error

    """
    #extract the parameters
    x_centre, y_centre, z_centre, radius = parameters

    #use numpy's sqrt function here, which works by element on arrays
    distance_from_centre = numpy.sqrt((x_values - x_centre)**2 +
                                      (y_values - y_centre)**2 +
                                      (z_values - z_centre)**2)

    return distance_from_centre - radius



# -*- coding: utf-8 -*-

""" SmartLiver calibration interface. """

import logging
from typing import List, Tuple
import numpy as np
from scipy.optimize import least_squares

LOGGER = logging.getLogger(__name__)


# pylint: disable=invalid-name, no-member, unsubscriptable-object
# pylint: disable=too-many-locals, too-many-arguments, too-many-lines
# pylint: disable=too-many-branches

def quat_conjugate(q):
    """
    Obtains the conjugate of a quaternion.
    """
    assert len(q) == 4

    result = [q[0], -q[1], -q[2], -q[3]]

    return np.asarray(result)


def quat_multiply(q, r):
    """
    Calculates the quaternion product n, for two given quaternions q and r:
    n = q x r
    """
    assert len(q) == 4
    assert len(r) == 4
    n0 = r[0] * q[0] - r[1] * q[1] - r[2] * q[2] - r[3] * q[3]
    n1 = r[0] * q[1] + r[1] * q[0] - r[2] * q[3] + r[3] * q[2]
    n2 = r[0] * q[2] + r[1] * q[3] + r[2] * q[0] - r[3] * q[1]
    n3 = r[0] * q[3] - r[1] * q[2] + r[2] * q[1] + r[3] * q[0]

    result = [n0, n1, n2, n3]

    return np.asarray(result)


def quat_to_rotm(q):
    """
    Get the corresponding rotation matrix of a quaternion.
    """
    assert len(q) == 4

    qn = q / np.linalg.norm(q)
    s = qn[0]
    x = qn[1]
    y = qn[2]
    z = qn[3]
    M = [[s**2 + x**2 - y**2 - z**2, 2*(x*y - s*z), 2*(x*z + s*y)],
         [2*(x*y + s*z), s**2 - x**2 + y**2 - z**2, 2*(y*z - s*x)],
         [2*(x*z - s*y), 2*(y*z + s*x), s**2 - x**2 - y**2 + z**2]]

    return np.asarray(M)


def rotm_to_quat_precise(M):
    """
    Get the corresponding quaternion of a rotation matrix.
    Assuming the rotation matrix is orthonormal.
    """
    q = np.empty((4,))
    tr = np.trace(M)
    if tr > 0:
        q[0] = tr + 1.0
        q[3] = M[1, 0] - M[0, 1]
        q[2] = M[0, 2] - M[2, 0]
        q[1] = M[2, 1] - M[1, 2]
    else:
        i, j, k = 0, 1, 2
        if M[1, 1] > M[0, 0]:
            i, j, k = 1, 2, 0
        if M[2, 2] > M[i, i]:
            i, j, k = 2, 0, 1
        tr = M[i, i] - (M[j, j] + M[k, k])
        q[i] = tr + 1.0
        q[j] = M[i, j] + M[j, i]
        q[k] = M[k, i] + M[i, k]
        q[3] = M[k, j] - M[j, k]
        q = q[[3, 0, 1, 2]]

    q *= 0.5 / np.sqrt(tr + 1.0)

    if q[0] < 0.0:
        np.negative(q, q)

    return q


def rotm_to_quat(M):
    """
    Get the corresponding quaternion of a rotation matrix.
    Assuming the rotation matrix is not strictly orthonormal.
    """
    m00 = M[0, 0]
    m01 = M[0, 1]
    m02 = M[0, 2]
    m10 = M[1, 0]
    m11 = M[1, 1]
    m12 = M[1, 2]
    m20 = M[2, 0]
    m21 = M[2, 1]
    m22 = M[2, 2]

    # symmetric matrix K
    K = np.array([[m00-m11-m22, 0.0, 0.0, 0.0],
                  [m01+m10, m11-m00-m22, 0.0, 0.0],
                  [m02+m20, m12+m21, m22-m00-m11, 0.0],
                  [m21-m12, m02-m20, m10-m01, m00+m11+m22]])
    K /= 3.0

    # quaternion is eigenvector of K that corresponds to largest eigenvalue
    w, V = np.linalg.eigh(K)
    q = V[[3, 0, 1, 2], np.argmax(w)]

    if q[0] < 0.0:
        np.negative(q, q)

    return q


def rvec_to_quat(rvec):
    """
    Get the corresponding quaternion of a rotation vector.
    """
    rad = np.linalg.norm(rvec)
    axis = rvec/rad
    q = np.concatenate([[np.cos(rad/2)], np.sin(rad/2) * axis])

    return q


def to_one_hemisphere(quaternions):
    """
    Transform a group of quaternions to one hemisphere.
    """
    new_quaternions = np.zeros(np.shape(quaternions))

    sum_1 = np.sum(np.abs(quaternions[:, 0]))
    sum_2 = np.sum(np.abs(quaternions[:, 1]))
    sum_3 = np.sum(np.abs(quaternions[:, 2]))
    sum_4 = np.sum(np.abs(quaternions[:, 3]))

    max_sum = sum_1
    flag = 0

    if sum_2 > max_sum:
        max_sum = sum_2
        flag = 1

    if sum_3 > max_sum:
        max_sum = sum_3
        flag = 2

    if sum_4 > max_sum:
        flag = 3

    number = np.shape(quaternions)[0]

    for i in range(0, number):
        if quaternions[i, flag] > 0:
            new_quaternions[i] = quaternions[i]
        else:
            new_quaternions[i] = -quaternions[i]

    return new_quaternions


def solve_2quaternions(qx, q_A, q_B):
    """
    Provide cost function for least-square optimisation
    to solve quaternions qx1 and qx2 in: q_B = qx1 * q_A * qx2

    :param qx: 1x8 vector of 2 quaternions
    :param q_A: Nx4 matrices of N quaternions
    :param q_B: Nx4 matrices of N quaternions
    :return:
    """
    if np.shape(q_A) != np.shape(q_B):
        raise AssertionError("q_A and q_B must have the same shape.")

    number_of_frames = np.shape(q_A)[0]

    f = []

    qx_1 = qx[0:4]
    qx_2 = qx[4:8]

    qx_1 = qx_1/np.linalg.norm(qx_1)
    qx_2 = qx_2/np.linalg.norm(qx_2)

    for i in range(0, number_of_frames):
        f.append(quat_multiply(quat_multiply(qx_1, q_A[i]), qx_2) - q_B[i])

    f = np.ndarray.flatten(np.asarray(f))

    return f


def solve_2translations(q_handeye,
                        quat_model2hand_array,
                        trans_model2hand_array,
                        trans_extrinsics_array):
    """
    Solve translations t_handeye and t_pattern2marker.
    translations = [t_handeye t_pattern2marker]

    :param q_handeye: 1x4 matrix quaternion
    :param quat_model2hand_array: Nx4 matrix of N quaternions
    :param trans_model2hand_array: Nx3 matrices of N translations
    :param trans_extrinsics_array: Nx3 matrices of extrinsic translation vectors
    :return:
    """
    if np.shape(quat_model2hand_array)[0] \
            != np.shape(trans_model2hand_array)[0]:
        raise AssertionError("Wrong dimensions in solve_2translations().")
    if np.shape(trans_model2hand_array) != np.shape(trans_extrinsics_array):
        raise AssertionError("Wrong dimensions in solve_2translations().")

    number_of_frames = np.shape(quat_model2hand_array)[0]

    A = np.zeros((3 * number_of_frames, 6))
    B = np.zeros((3 * number_of_frames, 1))

    for i in range(0, number_of_frames):
        qHEqMH = quat_multiply(q_handeye, quat_model2hand_array[i])
        rHErMH = quat_to_rotm(qHEqMH)
        rHE = quat_to_rotm(q_handeye)

        A[i*3:i*3+3, :] = np.hstack([np.identity(3), rHErMH])
        b_i = np.transpose(trans_extrinsics_array[i, :])\
            - np.matmul(rHE, np.transpose(trans_model2hand_array[i, :]))
        B[i*3:i*3+3] = np.reshape(b_i, (3, 1))

    A_T = np.transpose(A)
    translations = np.linalg.inv(A_T @ A) @ A_T @ B
    translations = np.transpose(translations)

    return translations.flatten()


def set_model2hand_arrays(calibration_tracking_array: List,
                          device_tracking_array: List,
                          use_quaternions=False) \
                            -> Tuple[np.ndarray, np.ndarray]:
    """ Set the model-to-hand quaternion and translation arrays
    from tracking data.

    :param calibration_tracking_array: Array of tracking data for
    calibration target
    :type calibration_tracking_array: List of tracking data
    :param device_tracking_array: Array of tracking data for
    device (e.g. camera)
    :type device_tracking_array: List of tracking data

    :return: quaternion model to hand array and translation model to hand
    array
    :rtype: np.ndarray, np.ndarray
    """

    if len(calibration_tracking_array) != len(device_tracking_array):
        raise ValueError('Calibration target and device tracking array \
                            should be the same size')

    number_of_frames = len(calibration_tracking_array)

    quat_model2hand_array = np.zeros((number_of_frames, 4))
    trans_model2hand_array = np.zeros((number_of_frames, 3))

    for i in range(number_of_frames):
        if use_quaternions:
            quat_model = \
                calibration_tracking_array[i][0, 0:4]
            trans_model = \
                calibration_tracking_array[i][0, 4:7]
            quat_hand = \
                device_tracking_array[i][0, 0:4]
            trans_hand = \
                device_tracking_array[i][0, 4:7]

            quat_model2hand = quat_multiply(quat_conjugate(quat_hand),
                                            quat_model)
            q_translation = np.append([0], (trans_model - trans_hand))
            quat_hand_conj = quat_conjugate(quat_hand)

            trans_model2hand = quat_multiply(quat_hand_conj,
                                             quat_multiply(q_translation,
                                                           quat_hand))
            trans_model2hand = trans_model2hand[1:4]
        else:
            tracking_model = calibration_tracking_array[i]
            tracking_hand = device_tracking_array[i]
            model_to_hand = np.linalg.inv(tracking_hand) @ tracking_model

            quat_model2hand = rotm_to_quat(model_to_hand[0:3, 0:3])
            trans_model2hand = model_to_hand[0:3, 3]

        quat_model2hand_array[i] = quat_model2hand
        trans_model2hand_array[i] = trans_model2hand

    quat_model2hand_array = \
        to_one_hemisphere(quat_model2hand_array)

    return quat_model2hand_array, trans_model2hand_array


def handeye_optimisation(quat_extrinsics_array: np.ndarray,
                         trans_extrinsics_array: np.ndarray,
                         quat_model2hand_array: np.ndarray,
                         trans_model2hand_array: np.ndarray) ->  \
                          Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Solve handeye and pattern-to-marker transformations.

    :param quat_extrinsics_array: An array of quaternions representing the
        rotations of the camera extrinsics.
    :type quat_extrinsics_array: np.ndarray
    :param trans_extrinsics_array:  Array of the translation vectors of the
        camera extrinsics.
    :type trans_extrinsics_array: np.ndarray
    :param quat_model2hand_array: An array of model to hand quaternions.
        :type quat_model2hand_array: np.ndarray
    :param quat_model2hand_array: An array of model to hand
        translations arrays.
        :type quat_model2hand_array: np.ndarray
    :return: rotations in quaternion form and translations of the handeye
        pattern-to-marker transformation.
    :rtype: np.ndarray
    """

    # self.quat_model2hand_array should already be in one hemisphere.

    quat_extrinsics_array = to_one_hemisphere(quat_extrinsics_array)

    qx_0 = [1, 0, 0, 0, 1, 0, 0, 0]
    lb = [-1, -1, -1, -1, -1, -1, -1, -1]
    ub = [1, 1, 1, 1, 1, 1, 1, 1]

    op_result = least_squares(solve_2quaternions, qx_0,
                              bounds=(lb, ub),
                              args=(quat_model2hand_array,
                                    quat_extrinsics_array)
                              )

    q_handeye = op_result.x[0:4]
    q_pattern2marker = op_result.x[4:8]

    q_handeye = q_handeye / np.linalg.norm(q_handeye)
    q_pattern2marker = q_pattern2marker / np.linalg.norm(q_pattern2marker)

    translations = solve_2translations(q_handeye,
                                       quat_model2hand_array,
                                       trans_model2hand_array,
                                       trans_extrinsics_array)

    t_handeye = translations[0:3]
    t_pattern2marker = translations[3:6]

    return q_handeye, t_handeye, q_pattern2marker, t_pattern2marker


def handeye_calibration(rvecs: List[np.ndarray], tvecs: List[np.ndarray],
                        quat_model2hand_array: np.ndarray,
                        trans_model2hand_array: np.ndarray) \
                             -> Tuple[np.ndarray, np.ndarray]:
    """Solve for the handeye transformation, as well as the transformation
     from the pattern to the markers on the model.

    :param rvecs: Array of rotation vectors
    :type rvecs: List[np.ndarray]
    :param tvecs: Array of translation vectors
    :type tvecs: List[np.ndarray]
    :return: [description]
    :rtype: np.ndarray
    """

    number_of_frames = len(rvecs)
    quat_extrinsics_array = np.zeros((number_of_frames, 4))

    for i in range(number_of_frames):
        quat_extrinsics_array[i] = rvec_to_quat(rvecs[i].flatten())

    trans_extrinsics_array = np.reshape(tvecs, (number_of_frames, 3))

    q_handeye, t_handeye, q_pattern2marker, t_pattern2marker = \
        handeye_optimisation(quat_extrinsics_array, trans_extrinsics_array,
                             quat_model2hand_array, trans_model2hand_array)

    handeye_matrix = np.eye(4)
    handeye_matrix[0:3, 0:3] = quat_to_rotm(q_handeye)
    handeye_matrix[0:3, 3] = t_handeye

    pattern2marker_matrix = np.eye(4)
    pattern2marker_matrix[0:3, 0:3] = quat_to_rotm(q_pattern2marker)
    pattern2marker_matrix[0:3, 3] = t_pattern2marker

    return handeye_matrix, pattern2marker_matrix



"""Tests for command line application """
from sksurgerycalibration.ui.pivot_calibration_command_line import main

def test_cl_no_config():
    """ Run command line app with no config file """
    main(['-i', 'tests/data/PivotCalibration/'])


def test_cl_ransac_config():
    """ Run command line app with ransac in config """
    main(['-i', 'tests/data/PivotCalibration/', '-c',
          'config/ransac_conf.json'])



# -*- coding: utf-8 -*-

""" Video calibration metrics. """

# pylint: disable=too-many-arguments

import logging
from typing import List
import numpy as np
import cv2
import sksurgeryopencvpython as cvpy
import sksurgerycore.transforms.matrix as mu
import sksurgerycalibration.video.video_calibration_utils as vu

LOGGER = logging.getLogger(__name__)


def compute_stereo_2d_err(l2r_rmat,
                          l2r_tvec,
                          left_object_points,
                          left_image_points,
                          left_camera_matrix,
                          left_distortion,
                          right_object_points,
                          right_image_points,
                          right_camera_matrix,
                          right_distortion,
                          left_rvecs,
                          left_tvecs):
    """
    Function to compute stereo SSE re-projection error, over multiple views.

    :param l2r_rmat: [3x3] ndarray, rotation for l2r transform
    :param l2r_tvec: [3x1] ndarray, translation for l2r transform
    :param left_object_points: Vector of Vector of 1x3 of type float32
    :param left_image_points: Vector of Vector of 1x2 of type float32
    :param left_camera_matrix: [3x3] ndarray
    :param left_distortion: [1x5] ndarray
    :param right_object_points: Vector of Vector of 1x3 of type float32
    :param right_image_points: Vector of Vector of 1x2 of type float32
    :param right_camera_matrix: [3x3] ndarray
    :param right_distortion: [1x5] ndarray
    :param left_rvecs: Vector of [3x1] ndarray, Rodrigues rotations, left camera
    :param left_tvecs: Vector of [3x1] ndarray, translations, left camera
    :return: SSE re-reprojection error, number_samples
    """
    left_to_right = mu.construct_rigid_transformation(l2r_rmat, l2r_tvec)

    lse = 0
    rse = 0
    number_of_samples = 0
    number_of_frames = len(left_object_points)

    for i in range(0, number_of_frames):

        projected_left, _ = cv2.projectPoints(left_object_points[i],
                                              left_rvecs[i],
                                              left_tvecs[i],
                                              left_camera_matrix,
                                              left_distortion)

        world_to_left_camera = vu.extrinsic_vecs_to_matrix(left_rvecs[i],
                                                           left_tvecs[i])

        world_to_right_camera = np.matmul(left_to_right, world_to_left_camera)

        rvec, tvec = vu.extrinsic_matrix_to_vecs(world_to_right_camera)

        projected_right, _ = cv2.projectPoints(right_object_points[i],
                                               rvec,
                                               tvec,
                                               right_camera_matrix,
                                               right_distortion)

        diff_left = left_image_points[i] - projected_left

        lse = lse + np.sum(np.square(diff_left))

        diff_right = right_image_points[i] - projected_right
        rse = rse + np.sum(np.square(diff_right))
        number_of_samples = number_of_samples \
            + len(left_image_points[i]) \
            + len(right_image_points[i])

    LOGGER.debug("Stereo RMS reprojection: left sse=%s, right sse=%s, num=%s",
                 str(lse), str(rse), str(number_of_samples))
    return lse + rse, number_of_samples


def compute_stereo_3d_error(l2r_rmat,
                            l2r_tvec,
                            common_object_points,
                            common_left_image_points,
                            left_camera_matrix,
                            left_distortion,
                            common_right_image_points,
                            right_camera_matrix,
                            right_distortion,
                            left_rvecs,
                            left_tvecs):
    """
    Function to compute stereo SSE reconstruction error over multiple views.

    :param l2r_rmat: [3x3] ndarray, rotation for l2r transform
    :param l2r_tvec: [3x1] ndarray, translation for l2r transform
    :param common_object_points: Vector of Vector of 1x3 of type float32
    :param common_left_image_points: Vector of Vector of 1x2 of type float32
    :param left_camera_matrix: [3x3] ndarray
    :param left_distortion: [1x5] ndarray
    :param common_right_image_points: Vector of Vector of 1x2 of type float32
    :param right_camera_matrix: [3x3] ndarray
    :param right_distortion: [1x5] ndarray
    :param left_rvecs: Vector of [3x1] ndarray, Rodrigues rotations, left camera
    :param left_tvecs: Vector of [3x1] ndarray, translations, left camera
    :return: SSE re-reprojection error, number_samples
    """
    sse = 0
    number_of_samples = 0
    number_of_frames = len(common_object_points)

    for i in range(0, number_of_frames):

        model_points = np.reshape(common_object_points[i], (-1, 3))

        left_undistorted = \
            cv2.undistortPoints(common_left_image_points[i],
                                left_camera_matrix,
                                left_distortion, None, left_camera_matrix)
        right_undistorted = \
            cv2.undistortPoints(common_right_image_points[i],
                                right_camera_matrix,
                                right_distortion, None, right_camera_matrix)

        # Triangulate using OpenCV
#        l2r_mat = mu.construct_rigid_transformation(l2r_rmat, l2r_tvec)
#        p_l = np.zeros((3, 4))
#        p_l[:, :-1] = left_camera_matrix
#        p_r = np.zeros((3, 4))
#        p_r[:, :-1] = right_camera_matrix
#        p_l = np.matmul(p_l, np.eye(4))
#        p_r = np.matmul(p_r, l2r_mat)
#        triangulated_cv = cv2.triangulatePoints(p_l,
#                                                p_r,
#                                                left_undistorted,
#                                                right_undistorted)

        # convert from Mx1x2 to Mx2
        left_undistorted = np.reshape(left_undistorted, (-1, 2))
        right_undistorted = np.reshape(right_undistorted, (-1, 2))

        image_points = np.zeros((left_undistorted.shape[0], 4))
        image_points[:, 0:2] = left_undistorted
        image_points[:, 2:4] = right_undistorted

        #pylint:disable=c-extension-no-member
        triangulated = cvpy.triangulate_points_using_hartley(
            image_points,
            left_camera_matrix,
            right_camera_matrix,
            l2r_rmat,
            l2r_tvec)

        # Triangulated points, are with respect to left camera.
        # Need to map back to model (chessboard space) for comparison.
        # Or, map chessboard points into left-camera space. Chose latter.
        rmat = (cv2.Rodrigues(left_rvecs[i]))[0]
        rotated = np.matmul(rmat, np.transpose(model_points))
        translated = rotated + left_tvecs[i]  # uses broadcasting
        transformed = np.transpose(translated)

        # Now compute squared error
        diff = triangulated - transformed
        squared = np.square(diff)
        sum_square = np.sum(squared)
        sse = sse + sum_square
        number_of_samples = number_of_samples + len(common_left_image_points[i])

    LOGGER.debug("Stereo RMS reconstruction: sse=%s, num=%s",
                 str(sse), str(number_of_samples))
    return sse, number_of_samples


def compute_mono_2d_err(object_points,
                        image_points,
                        rvecs,
                        tvecs,
                        camera_matrix,
                        distortion):
    """
    Function to compute mono RMS reprojection error over multiple views.

    :param object_points: Vector of Vector of 1x3 of type float32
    :param image_points: Vector of Vector of 1x2 of type float32
    :param rvecs: Vector of [3x1] ndarray, Rodrigues rotations for each camera
    :param tvecs: Vector of [3x1] ndarray, translations for each camera
    :param camera_matrix: [3x3] ndarray
    :param distortion: [1x5] ndarray
    :return: SSE re-reprojection error, number_samples
    """
    sse = 0
    number_of_samples = 0
    number_of_frames = len(object_points)

    for i in range(0, number_of_frames):

        projected, _ = cv2.projectPoints(object_points[i],
                                         rvecs[i],
                                         tvecs[i],
                                         camera_matrix,
                                         distortion)

        diff = image_points[i] - projected
        squared = np.square(diff)
        sum_square = np.sum(squared)
        sse = sse + sum_square
        number_of_samples = number_of_samples + len(image_points[i])

    LOGGER.debug("Mono RMS reprojection: sse=%s, num=%s",
                 str(sse), str(number_of_samples))
    return sse, number_of_samples


def compute_mono_3d_err(ids,
                        object_points,
                        image_points,
                        rvecs,
                        tvecs,
                        camera_matrix,
                        distortion):
    """
    Function to compute mono RMS reconstruction error over multiple views.

    Here, to triangulate, we take the i^th camera as left camera, and
    the i+1^th camera as the right camera, compute l2r, and triangulate.

    :param ids: Vector of ndarray of integer point ids
    :param object_points: Vector of Vector of 1x3 of type float32
    :param image_points: Vector of Vector of 1x2 of type float32
    :param rvecs: Vector of [3x1] ndarray, Rodrigues rotations for each camera
    :param tvecs: Vector of [3x1] ndarray, translations for each camera
    :param camera_matrix: [3x3] ndarray
    :param distortion: [1x5] ndarray
    :return: SSE re-reprojection error, number_samples
    """

    sse = 0
    number_of_samples = 0
    number_of_frames = len(object_points)

    # We are going to triangulate between a frame and the next frame.
    for i in range(0, number_of_frames):

        j = i + 1
        if j == number_of_frames:
            j = 0

        _, common_object_points, common_left_image_points, \
            common_right_image_points = \
            vu.filter_common_points_per_image(ids[i],
                                              object_points[i],
                                              image_points[i],
                                              ids[j],
                                              image_points[j],
                                              10)

        left_camera_to_world = vu.extrinsic_vecs_to_matrix(rvecs[i],
                                                           tvecs[i])

        right_camera_to_world = vu.extrinsic_vecs_to_matrix(rvecs[j],
                                                            tvecs[j])
        left_to_right = np.matmul(right_camera_to_world,
                                  np.linalg.inv(left_camera_to_world))
        l2r_rmat = left_to_right[0:3, 0:3]
        l2r_tvec = left_to_right[0:3, 3]

        c_obj = [common_object_points]
        c_li = [common_left_image_points]
        c_ri = [common_right_image_points]
        r_v = [rvecs[i]]
        t_v = [tvecs[i]]
        err, samp = compute_stereo_3d_error(l2r_rmat,
                                            l2r_tvec,
                                            c_obj,
                                            c_li,
                                            camera_matrix,
                                            distortion,
                                            c_ri,
                                            camera_matrix,
                                            distortion,
                                            r_v,
                                            t_v)
        sse = sse + err
        number_of_samples = number_of_samples + samp

    LOGGER.debug("Mono RMS reconstruction: sse=%s, num=%s",
                 str(sse), str(number_of_samples))

    return sse, number_of_samples


def compute_mono_2d_err_handeye(model_points: List,
                                image_points: List,
                                camera_matrix: np.ndarray,
                                camera_distortion: np.ndarray,
                                hand_tracking_array: List,
                                model_tracking_array: List,
                                handeye_matrix: np.ndarray,
                                pattern2marker_matrix: np.ndarray):
    """Function to compute mono SSE reprojection error

    :param model_points: Vector of Vector of 1x3 float32
    :type model_points: List
    :param image_points: Vector of Vector of 1x2 float32
    :type image_points: List
    :param camera_matrix: Camera intrinsic matrix
    :type camera_matrix: np.ndarray
    :param camera_distortion: Camera distortion coefficients
    :type camera_distortion: np.ndarray
    :param hand_tracking_array:
    Vector of 4x4 tracking matrices for camera (hand)
    :type hand_tracking_array: List
    :param model_tracking_array:
    Vector of 4x4 tracking matrices for calibration model
    :type model_tracking_array: List
    :param handeye_matrix: Handeye matrix
    :type handeye_matrix: np.ndarray
    :param pattern2marker_matrix: Pattern to marker matrix
    :type pattern2marker_matrix: np.ndarray
    :return: SSE reprojection error, number of samples
    :rtype: float, float
    """

    number_of_frames = len(model_points)

    rvecs = []
    tvecs = []

    # Construct rvec/tvec array taking into account handeye calibration.
    # Then the rest of the calculation can use 'normal' compute_mono_2d_err()
    for i in range(number_of_frames):
        pattern_to_camera = \
            handeye_matrix @ np.linalg.inv(hand_tracking_array[i]) @ \
                model_tracking_array[i] @ pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(pattern_to_camera)
        rvecs.append(rvec)
        tvecs.append(tvec)

    sse, number_of_samples = compute_mono_2d_err(model_points,
                                                 image_points,
                                                 rvecs,
                                                 tvecs,
                                                 camera_matrix,
                                                 camera_distortion)

    LOGGER.debug("Mono Handeye RMS Reprojection: sse=%s, num=%s",
                 str(sse), str(number_of_samples))

    return sse, number_of_samples


def compute_mono_3d_err_handeye(ids: List,
                                model_points: List,
                                image_points: List,
                                camera_matrix: np.ndarray,
                                camera_distortion: np.ndarray,
                                hand_tracking_array: List,
                                model_tracking_array: List,
                                handeye_matrix: np.ndarray,
                                pattern2marker_matrix: np.ndarray):
    """Function to compute mono SSE reconstruction error. Calculates new
    rvec/tvec values for pattern_to_camera based on handeye calibration and
    then calls compute_mono_3d_err().

    :param ids: Vector of ndarray of integer point ids
    :type ids: List
    :param model_points: Vector of Vector of 1x3 float32
    :type model_points: List
    :param image_points: Vector of Vector of 1x2 float32
    :type image_points: List
    :param camera_matrix: Camera intrinsic matrix
    :type camera_matrix: np.ndarray
    :param camera_distortion: Camera distortion coefficients
    :type camera_distortion: np.ndarray
    :param hand_tracking_array:
    Vector of 4x4 tracking matrices for camera (hand)
    :type hand_tracking_array: List
    :param model_tracking_array:
    Vector of 4x4 tracking matrices for calibration model
    :type model_tracking_array: List
    :param handeye_matrix: Handeye matrix
    :type handeye_matrix: np.ndarray
    :param pattern2marker_matrix: Pattern to marker matrix
    :type pattern2marker_matrix: np.ndarray
    :return: SSE reprojection error, number of samples
    :rtype: float, float
    """

    number_of_frames = len(model_points)

    rvecs = []
    tvecs = []

    # Construct rvec/tvec array taking into account handeye calibration.
    # Then the rest of the calculation can use 'normal' compute_mono_3d_err()
    for i in range(number_of_frames):
        pattern_to_camera = \
            handeye_matrix @ np.linalg.inv(hand_tracking_array[i]) @ \
                model_tracking_array[i] @ pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(pattern_to_camera)
        rvecs.append(rvec)
        tvecs.append(tvec)

    sse, number_of_samples = compute_mono_3d_err(ids,
                                                 model_points,
                                                 image_points,
                                                 rvecs,
                                                 tvecs,
                                                 camera_matrix,
                                                 camera_distortion)

    return sse, number_of_samples


def compute_stereo_2d_err_handeye(common_object_points: List,
                                  left_image_points: List,
                                  left_camera_matrix: np.ndarray,
                                  left_distortion: np.ndarray,
                                  right_image_points: List,
                                  right_camera_matrix: np.ndarray,
                                  right_distortion: np.ndarray,
                                  hand_tracking_array: List,
                                  model_tracking_array: List,
                                  left_handeye_matrix: np.ndarray,
                                  left_pattern2marker_matrix: np.ndarray,
                                  right_handeye_matrix: np.ndarray,
                                  right_pattern2marker_matrix: np.ndarray):
    """Function to compute stereo SSE reprojection error, taking into account
    handeye calibration.

    :param common_object_points: Vector of Vector of 1x3 float32
    :type common_object_points: List
    :param left_image_points: Vector of Vector of 1x2 float32
    :type left_image_points: List
    :param left_camera_matrix: Left camera matrix
    :type left_camera_matrix: np.ndarray
    :param left_distortion: Left camera distortion coefficients
    :type left_distortion: np.ndarray
    :param right_image_points: Vector of Vector of 1x2 float32
    :type right_image_points: List
    :param right_camera_matrix: Right camera matrix
    :type right_camera_matrix: np.ndarray
    :param right_distortion: Right camera distortion coefficients
    :type right_distortion: np.ndarray
    :param hand_tracking_array:
    Vector of 4x4 tracking matrices for camera (hand)
    :type hand_tracking_array: List
    :param model_tracking_array:
    Vector of 4x4 tracking matrices for calibration model
    :type model_tracking_array: List
    :param left_handeye_matrix: Left handeye transform matrix
    :type left_handeye_matrix: np.ndarray
    :param left_pattern2marker_matrix: Left pattern to marker transform matrix
    :type left_pattern2marker_matrix: np.ndarray
    :param right_handeye_matrix: Right handeye transform matrix
    :type right_handeye_matrix: np.ndarray
    :param right_pattern2marker_matrix: Right pattern to marker transform matrix
    :type right_pattern2marker_matrix: np.ndarray
    :return: SSE reprojection error, number of samples
    :rtype: float, float
    """

    lse, l_samples = compute_mono_2d_err_handeye(common_object_points,
                                                 left_image_points,
                                                 left_camera_matrix,
                                                 left_distortion,
                                                 hand_tracking_array,
                                                 model_tracking_array,
                                                 left_handeye_matrix,
                                                 left_pattern2marker_matrix)

    rse, r_samples = compute_mono_2d_err_handeye(common_object_points,
                                                 right_image_points,
                                                 right_camera_matrix,
                                                 right_distortion,
                                                 hand_tracking_array,
                                                 model_tracking_array,
                                                 right_handeye_matrix,
                                                 right_pattern2marker_matrix)

    return lse + rse, l_samples + r_samples


def compute_stereo_3d_err_handeye(l2r_rmat: np.ndarray,
                                  l2r_tvec: np.ndarray,
                                  common_object_points: List,
                                  common_left_image_points: List,
                                  left_camera_matrix: np.ndarray,
                                  left_distortion: np.ndarray,
                                  common_right_image_points: List,
                                  right_camera_matrix: np.ndarray,
                                  right_distortion: np.ndarray,
                                  hand_tracking_array: List,
                                  model_tracking_array: List,
                                  left_handeye_matrix: np.ndarray,
                                  left_pattern2marker_matrix: np.ndarray):

    """Function to compute stereo SSE reconstruction error, taking into account
    handeye calibration.

    :param l2r_rmat: Rotation for l2r transform
    :type l2r_rmat: np.ndarray
    :param l2r_tvec: Translation for l2r transform
    :type l2r_tvec: np.ndarray
    :param common_object_points: Vector of Vector of 1x3 float32
    :type common_object_points: List
    :param common_left_image_points: Vector of Vector of 1x2 float32
    :type common_left_image_points: List
    :param left_camera_matrix: Left camera matrix
    :type left_camera_matrix: np.ndarray
    :param left_distortion: Left camera distortion coefficients
    :type left_distortion: np.ndarray
    :param common_right_image_points: Vector of Vector of 1x2 float32
    :type common_right_image_points: List
    :param right_camera_matrix: Right camera matrix
    :type right_camera_matrix: np.ndarray
    :param right_distortion: Right camera distortion coefficients
    :type right_distortion: np.ndarray
    :param hand_tracking_array:
    Vector of 4x4 tracking matrices for camera (hand)
    :type hand_tracking_array: List
    :param model_tracking_array:
    Vector of 4x4 tracking matrices for calibration model
    :type model_tracking_array: List
    :param left_handeye_matrix: Left handeye transform matrix
    :type left_handeye_matrix: np.ndarray
    :param left_pattern2marker_matrix: Left pattern to marker transform matrix
    :type left_pattern2marker_matrix: np.ndarray
    :return: SSE reconstruction error, number of samples
    :rtype: float, float
    """

    number_of_frames = len(common_object_points)
    left_rvecs = []
    left_tvecs = []

    # Construct rvec/tvec array taking into account handeye calibration.
    # Then the rest of the calculation can use 'normal' compute_stereo_3d_err()
    for i in range(number_of_frames):

        pattern_to_left_camera = \
            left_handeye_matrix @ np.linalg.inv(hand_tracking_array[i]) @ \
                 model_tracking_array[i] @ left_pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(pattern_to_left_camera)
        left_rvecs.append(rvec)
        left_tvecs.append(tvec)

    sse, number_of_samples = compute_stereo_3d_error(l2r_rmat,
                                                     l2r_tvec,
                                                     common_object_points,
                                                     common_left_image_points,
                                                     left_camera_matrix,
                                                     left_distortion,
                                                     common_right_image_points,
                                                     right_camera_matrix,
                                                     right_distortion,
                                                     left_rvecs,
                                                     left_tvecs)

    LOGGER.debug("Stereo Handeye RMS reconstruction: sse=%s, num=%s",
                 str(sse), str(number_of_samples))

    return sse, number_of_samples



# -*- coding: utf-8 -*-

# pylint: disable=unused-import, superfluous-parens, line-too-long, missing-module-docstring, unused-variable, missing-function-docstring, invalid-name

import glob
import pytest
import numpy as np
import cv2
import sksurgeryimage.calibration.charuco as ch
import sksurgeryimage.calibration.charuco_plus_chessboard_point_detector as pd
import sksurgerycalibration.video.video_calibration_driver_stereo as sc


def test_stereo_davinci():

    left_images = []
    files = glob.glob('tests/data/ChAruco_LR_frames_Steve_Axis_Tests/ExtractedFrames_L/*.jpg')
    files.sort()
    for file in files:
        image = cv2.imread(file)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        print("Loaded:" + str(file))
        left_images.append(image)
    assert(len(left_images) == 59)

    right_images = []
    files = glob.glob('tests/data/ChAruco_LR_frames_Steve_Axis_Tests/ExtractedFrames_R/*.jpg')
    files.sort()
    for file in files:
        image = cv2.imread(file)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        print("Loaded:" + str(file))
        right_images.append(image)
    assert (len(right_images) == 59)

    minimum_number_of_points_per_image = 50
    detector = pd.CharucoPlusChessboardPointDetector(error_if_no_chessboard=False) # Try to accept as many as possible.
    calibrator = sc.StereoVideoCalibrationDriver(detector, minimum_number_of_points_per_image)
    for i, _ in enumerate(left_images):
        try:
            number_of_points = calibrator.grab_data(left_images[i], right_images[i])
            if number_of_points < 2 * minimum_number_of_points_per_image:
                print("Image pair:" + str(i) + ", SKIPPED, due to not enough points")
        except ValueError as ve:
            print("Image pair:" + str(i) + ", FAILED, due to:" + str(ve))
        except TypeError as te:
            print("Image pair:" + str(i) + ", FAILED, due to:" + str(te))

    reproj_err, recon_err, params = calibrator.calibrate()
    print("Reproj:" + str(reproj_err))
    print("Recon:" + str(recon_err))
    assert reproj_err < 1.1
    assert recon_err < 6.4

    # Now try iterative.
    reference_image = ch.make_charuco_with_chessboard()
    reference_ids, object_pts, reference_pts = detector.get_points(reference_image)
    reproj_err, recon_err, params = calibrator.iterative_calibration(2,
                                                                     reference_ids,
                                                                     reference_pts,
                                                                     (reference_image.shape[1], reference_image.shape[0]))
    print("Reproj:" + str(reproj_err))
    print("Recon:" + str(recon_err))
    assert reproj_err < 0.9
    assert recon_err < 2.78
    calibrator.save_params('tests/output/ChAruco_LR_frames_Steve_Axis_Tests/params', '')
    calibrator.save_data('tests/output/ChAruco_LR_frames_Steve_Axis_Tests/data', '')



# -*- coding: utf-8 -*-

""" Various functions to help with IO. Not intended for 3rd party clients. """

import os
import glob

# pylint: disable=missing-function-docstring, invalid-name


def get_calib_prefix(file_prefix: str):

    prefix = 'calib'
    if file_prefix:
        prefix = file_prefix
    return prefix


def get_left_prefix(file_prefix: str):

    left_prefix = "calib.left"
    if file_prefix:
        left_prefix = file_prefix + ".left"
    return left_prefix


def get_right_prefix(file_prefix: str):

    right_prefix = "calib.right"
    if file_prefix:
        right_prefix = file_prefix + ".right"
    return right_prefix


def get_intrinsics_file_name(dir_name: str,
                             file_prefix: str):

    intrinsics_file = os.path.join(dir_name,
                                   get_calib_prefix(file_prefix) +
                                   ".intrinsics.txt")
    return intrinsics_file


def get_distortion_file_name(dir_name: str,
                             file_prefix: str):

    dist_coeff_file = os.path.join(dir_name,
                                   get_calib_prefix(file_prefix) +
                                   ".distortion.txt")
    return dist_coeff_file


def get_enumerated_file_name(dir_name: str,
                             file_prefix: str,
                             type_prefix: str,
                             view_number: str,
                             extension_wth_dot: str
                             ):

    # Keep in synch with _get_enumerated_file_glob
    file_name = \
        os.path.join(dir_name,
                     get_calib_prefix(file_prefix)
                     + "."
                     + type_prefix
                     + "."
                     + str(view_number) + extension_wth_dot)
    return file_name


def get_enumerated_file_glob(dir_name: str,
                             file_prefix: str,
                             type_prefix: str,
                             extension_wth_dot: str
                             ):

    # Keep in synch with _get_enumerated_file_name
    file_glob = \
        os.path.join(dir_name,
                     get_calib_prefix(file_prefix)
                     + "."
                     + type_prefix
                     + "."
                     + "*" + extension_wth_dot)

    return file_glob


def get_extrinsics_file_name(dir_name: str,
                             file_prefix: str,
                             view_number: int
                             ):

    extrinsics_file = get_enumerated_file_name(dir_name,
                                               file_prefix,
                                               "extrinsics",
                                               view_number,
                                               ".txt")
    return extrinsics_file


def get_extrinsic_file_names(dir_name: str,
                             file_prefix: str):

    files = get_filenames_by_glob_expr(dir_name,
                                       file_prefix,
                                       "extrinsics",
                                       ".txt")
    return files


def get_l2r_file_name(dir_name: str,
                      file_prefix: str):
    l2r_file = os.path.join(dir_name,
                            get_calib_prefix(file_prefix) + ".l2r.txt")
    return l2r_file


def get_essential_matrix_file_name(dir_name: str,
                                   file_prefix: str):
    ess_file = os.path.join(dir_name,
                            get_calib_prefix(file_prefix) + ".essential.txt")
    return ess_file


def get_fundamental_matrix_file_name(dir_name: str,
                                     file_prefix: str):
    fun_file = os.path.join(dir_name,
                            get_calib_prefix(file_prefix) + ".fundamental.txt")
    return fun_file


def get_images_file_name(dir_name: str,
                         file_prefix: str,
                         view_number: int
                         ):
    images_file = get_enumerated_file_name(dir_name,
                                           file_prefix,
                                           "images",
                                           view_number,
                                           ".png")
    return images_file


def get_ids_file_name(dir_name: str,
                      file_prefix: str,
                      view_number: int
                      ):
    ids_file = get_enumerated_file_name(dir_name,
                                        file_prefix,
                                        "ids",
                                        view_number,
                                        ".txt")
    return ids_file


def get_objectpoints_file_name(dir_name: str,
                               file_prefix: str,
                               view_number: int
                               ):
    object_points_file = get_enumerated_file_name(dir_name,
                                                  file_prefix,
                                                  "object_points",
                                                  view_number,
                                                  ".txt")
    return object_points_file


def get_imagepoints_file_name(dir_name: str,
                              file_prefix: str,
                              view_number: int
                              ):
    image_points_file = get_enumerated_file_name(dir_name,
                                                 file_prefix,
                                                 "image_points",
                                                 view_number,
                                                 ".txt")
    return image_points_file


def get_device_tracking_file_name(dir_name: str,
                                  file_prefix: str,
                                  view_number: int
                                  ):
    device_tracking = get_enumerated_file_name(dir_name,
                                               file_prefix,
                                               "device_tracking",
                                               view_number,
                                               ".txt")
    return device_tracking


def get_calibration_tracking_file_name(dir_name: str,
                                       file_prefix: str,
                                       view_number: int
                                       ):
    calibration_tracking = get_enumerated_file_name(dir_name,
                                                    file_prefix,
                                                    "calib_obj_tracking",
                                                    view_number,
                                                    ".txt")
    return calibration_tracking


def get_filenames_by_glob_expr(dir_name: str,
                               file_prefix: str,
                               type_prefix: str,
                               extension_with_dot: str
                               ):

    file_glob = get_enumerated_file_glob(dir_name,
                                         file_prefix,
                                         type_prefix,
                                         extension_with_dot)
    files = glob.glob(file_glob)
    files.sort()
    return files



# -*- coding: utf-8 -*-

""" Video Calibration functions, that wrap OpenCV functions mainly. """

import logging
from typing import List
import numpy as np
import cv2
import sksurgerycore.transforms.matrix as skcm
import sksurgerycalibration.video.video_calibration_utils as vu
import sksurgerycalibration.video.video_calibration_metrics as vm
import sksurgerycalibration.video.video_calibration_hand_eye as he

LOGGER = logging.getLogger(__name__)


def mono_video_calibration(object_points, image_points, image_size, flags=0):
    """
    Calibrates a video camera using Zhang's 2000 method, as implemented in
    OpenCV. We wrap it here, so we have a place to add extra validation code,
    and a space for documentation. The aim is to check everything before
    we pass it to OpenCV, and raise Exceptions consistently for any error
    we can detect before we pass it to OpenCV, as OpenCV just dies
    without throwing exceptions.

      - N = number of images
      - M = number of points for that image
      - rvecs = list of 1x3 Rodrigues rotation parameters
      - tvecs = list of 3x1 translation vectors
      - camera_matrix = [3x3] ndarray containing fx, fy, cx, cy
      - dist_coeffs = [1x5] ndarray, containing distortion coefficients

    :param object_points: Vector (N) of Vector (M) of 1x3 points of type float
    :param image_points: Vector (N) of Vector (M) of 1x2 points of type float
    :param image_size: (x, y) tuple, size in pixels, e.g. (1920, 1080)
    :param flags: OpenCV flags to pass to calibrateCamera().
    :return: rms, camera_matrix, dist_coeffs, rvecs, tvecs
    """
    if image_size[0] < 1:
        raise ValueError("Image width must be > 0.")
    if image_size[1] < 1:
        raise ValueError("Image height must be > 0.")
    if len(object_points) < 2:
        raise ValueError("Must have at least 2 sets of object points.")
    if len(image_points) < 2:
        raise ValueError("Must have at least 2 sets of image points.")
    if len(object_points) != len(image_points):
        raise ValueError("Image points and object points differ in length.")
    for i, _ in enumerate(object_points):
        if len(object_points[i]) < 3:
            raise ValueError(str(i) + ": Must have at least 3 object points.")
        if len(image_points[i]) < 3:
            raise ValueError(str(i) + ": Must have at least 3 image points.")
        if len(object_points[i]) != len(image_points[i]):
            raise ValueError(str(i) + ": Must have the same number of points.")

    _, camera_matrix, dist_coeffs, rvecs, tvecs \
        = cv2.calibrateCamera(object_points,
                              image_points,
                              image_size,
                              None, None,
                              flags=flags)

    # Recompute this, for consistency with stereo methods.
    # i.e. so we know what the calculation is exactly.
    sse, num = vm.compute_mono_2d_err(object_points,
                                      image_points,
                                      rvecs,
                                      tvecs,
                                      camera_matrix,
                                      dist_coeffs)
    mse = sse / num
    final_rms = np.sqrt(mse)

    return final_rms, camera_matrix, dist_coeffs, rvecs, tvecs


# pylint: disable=too-many-locals
def stereo_video_calibration(left_ids,
                             left_object_points,
                             left_image_points,
                             right_ids,
                             right_object_points,
                             right_image_points,
                             image_size,
                             flags=cv2.CALIB_USE_INTRINSIC_GUESS
                             ):
    """
    Default stereo calibration, using OpenCV methods.

    We wrap it here, so we have a place to add extra validation code,
    and a space for documentation. The aim is to check everything before
    we pass it to OpenCV, and raise Exceptions consistently for any error
    we can detect before we pass it to OpenCV.

    :param left_ids: Vector of ndarrays containing integer point ids.
    :param left_object_points: Vector of Vectors of 1x3 object points, float32
    :param left_image_points:  Vector of Vectors of 1x2 object points, float32
    :param right_ids: Vector of ndarrays containing integer point ids.
    :param right_object_points: Vector of Vectors of 1x3 object points, float32
    :param right_image_points: Vector of Vectors of 1x2 object points, float32
    :param image_size: (x, y) tuple, size in pixels, e.g. (1920, 1080)
    :param flags: OpenCV flags to pass to calibrateCamera().
    :return:
    """
    # Calibrate left, using all available points
    l_rms, l_c, l_d, l_rvecs, l_tvecs \
        = cv2.calibrateCamera(left_object_points,
                              left_image_points,
                              image_size,
                              None, None)

    # Calibrate right using all available points.
    r_rms, r_c, r_d, r_rvecs, r_tvecs \
        = cv2.calibrateCamera(right_object_points,
                              right_image_points,
                              image_size,
                              None, None)

    # But for stereo, OpenCV needs common points.
    _, common_object_points, common_left_image_points, \
        common_right_image_points \
        = vu.filter_common_points_all_images(left_ids,
                                             left_object_points,
                                             left_image_points,
                                             right_ids,
                                             right_image_points, 10)

    # First do stereo calibration, using fixed intrinsics.
    s_rms, l_c, l_d, r_c, r_d, \
        l2r_r, l2r_t, essential, fundamental = cv2.stereoCalibrate(
            common_object_points,
            common_left_image_points,
            common_right_image_points,
            l_c,
            l_d,
            r_c,
            r_d,
            image_size,
            flags=cv2.CALIB_USE_INTRINSIC_GUESS | cv2.CALIB_FIX_INTRINSIC)

    # Then do it again, using the passed in flags.
    s_rms, l_c, l_d, r_c, r_d, \
        l2r_r, l2r_t, essential, fundamental = cv2.stereoCalibrate(
            common_object_points,
            common_left_image_points,
            common_right_image_points,
            l_c,
            l_d,
            r_c,
            r_d,
            image_size,
            flags=flags)

    # And recompute rvecs and tvecs, consistently, given new l2r params.
    number_of_frames = len(left_object_points)
    left_to_right = skcm.construct_rigid_transformation(l2r_r, l2r_t)
    for i in range(0, number_of_frames):
        _, l_rvecs[i], l_tvecs[i] = cv2.solvePnP(
            common_object_points[i],
            common_left_image_points[i],
            l_c,
            l_d)
        left_chessboard_to_camera = \
            vu.extrinsic_vecs_to_matrix(l_rvecs[i], l_tvecs[i])
        right_chessboard_to_camera = \
            np.matmul(left_to_right, left_chessboard_to_camera)
        r_rvecs[i], r_tvecs[i] = \
            vu.extrinsic_matrix_to_vecs(right_chessboard_to_camera)

    # And recompute stereo projection error, given left camera and l2r.
    # We also use all points, not just common points, for comparison
    # with other methods outside of this function.
    sse, num_samples = \
        vm.compute_stereo_2d_err(l2r_r,
                                 l2r_t,
                                 common_object_points,
                                 common_left_image_points,
                                 l_c,
                                 l_d,
                                 common_object_points,
                                 common_right_image_points,
                                 r_c,
                                 r_d,
                                 l_rvecs,
                                 l_tvecs
                                 )
    mse = sse / num_samples
    s_reproj = np.sqrt(mse)

    sse, num_samples = \
        vm.compute_stereo_3d_error(l2r_r,
                                   l2r_t,
                                   common_object_points,
                                   common_left_image_points,
                                   l_c,
                                   l_d,
                                   common_right_image_points,
                                   r_c,
                                   r_d,
                                   l_rvecs,
                                   l_tvecs
                                   )

    mse = sse / num_samples
    s_recon = np.sqrt(mse)

    LOGGER.info("Stereo Calib: l=%s, r=%s, opencv=%s, proj=%s, recon=%s",
                str(l_rms), str(r_rms), str(s_rms), str(s_reproj), str(s_recon))

    return s_reproj, s_recon, \
        l_c, l_d, l_rvecs, l_tvecs, \
        r_c, r_d, r_rvecs, r_tvecs, \
        l2r_r, l2r_t, \
        essential, fundamental


#pylint:disable=too-many-arguments
def mono_handeye_calibration(object_points: List,
                             image_points: List,
                             ids: List,
                             camera_matrix: np.ndarray,
                             camera_distortion: np.ndarray,
                             device_tracking_array: List,
                             model_tracking_array: List,
                             rvecs: List[np.ndarray],
                             tvecs: List[np.ndarray],
                             quat_model2hand_array: List,
                             trans_model2hand_array: List):
    """Wrapper around handeye calibration functions and reprojection /
    reconstruction error metrics.

    :param object_points: Vector of Vectors of 1x3 object points, float32
    :type object_points: List
    :param image_points: Vector of Vectors of 1x2 object points, float32
    :type image_points: List
    :param ids: Vector of ndarrays containing integer point ids.
    :type ids: List
    :param camera_matrix: Camera intrinsic matrix
    :type camera_matrix: np.ndarray
    :param camera_distortion: Camera distortion coefficients
    :type camera_distortion: np.ndarray
    :param device_tracking_array: Tracking data for camera (hand)
    :type device_tracking_array: List
    :param model_tracking_array: Tracking data for calibration target
    :type model_tracking_array: List
    :param rvecs: Vector of 3x1 ndarray, Rodrigues rotations for each camera
    :type rvecs: List[np.ndarray]
    :param tvecs: Vector of [3x1] ndarray, translations for each camera
    :type tvecs: List[np.ndarray]
    :param quat_model2hand_array: Array of model to hand quaternions
    :type quat_model2hand_array: List
    :param trans_model2hand_array: Array of model to hand translaions
    :type trans_model2hand_array: List
    :return: Reprojection error, reconstruction error, handeye matrix,
    patter to marker matrix
    :rtype: float, float, np.ndarray, np.ndarray
    """
    handeye_matrix, pattern2marker_matrix =  \
        he.handeye_calibration(rvecs, tvecs, quat_model2hand_array,
                               trans_model2hand_array)

    sse, num_samples = vm.compute_mono_2d_err_handeye(object_points,
                                                      image_points,
                                                      camera_matrix,
                                                      camera_distortion,
                                                      device_tracking_array,
                                                      model_tracking_array,
                                                      handeye_matrix,
                                                      pattern2marker_matrix
                                                      )

    mse = sse / num_samples
    reproj_err = np.sqrt(mse)

    sse, num_samples = vm.compute_mono_3d_err_handeye(ids,
                                                      object_points,
                                                      image_points,
                                                      camera_matrix,
                                                      camera_distortion,
                                                      device_tracking_array,
                                                      model_tracking_array,
                                                      handeye_matrix,
                                                      pattern2marker_matrix)

    mse = sse / num_samples
    recon_err = np.sqrt(mse)

    return reproj_err, recon_err, handeye_matrix, pattern2marker_matrix


#pylint:disable=too-many-arguments
def stereo_handeye_calibration(l2r_rmat: np.ndarray,
                               l2r_tvec: np.ndarray,
                               left_ids: List,
                               left_object_points: List,
                               left_image_points: List,
                               right_ids: List,
                               right_image_points: List,
                               left_camera_matrix: np.ndarray,
                               left_camera_distortion: np.ndarray,
                               right_camera_matrix: np.ndarray,
                               right_camera_distortion: np.ndarray,
                               device_tracking_array: List,
                               calibration_tracking_array: List,
                               left_rvecs: List[np.ndarray],
                               left_tvecs: List[np.ndarray],
                               right_rvecs: List[np.ndarray],
                               right_tvecs: List[np.ndarray],
                               quat_model2hand_array: List,
                               trans_model2hand_array: List):
    """Wrapper around handeye calibration functions and reprojection /
    reconstruction error metrics.

    :param l2r_rmat: [3x3] ndarray, rotation for l2r transform
    :type l2r_rmat: np.ndarray
    :param l2r_tvec: [3x1] ndarray, translation for l2r transform
    :type l2r_tvec: np.ndarray
    :param left_ids: Vector of ndarrays containing integer point ids.
    :type left_ids: List
    :param left_object_points: Vector of Vector of 1x3 of type float32
    :type left_object_points: List
    :param left_image_points: Vector of Vector of 1x2 of type float32
    :type left_image_points: List
    :param right_ids: Vector of ndarrays containing integer point ids.
    :type right_ids: List
    :param right_image_points: Vector of Vector of 1x3 of type float32
    :type right_image_points: List
    :param left_camera_matrix: Camera intrinsic matrix
    :type left_camera_matrix: np.ndarray
    :param left_camera_distortion: Camera distortion coefficients
    :type left_camera_distortion: np.ndarray
    :param right_camera_matrix: Camera intrinsic matrix
    :type right_camera_matrix: np.ndarray
    :param right_camera_distortion: Camera distortion coefficients
    :type right_camera_distortion: np.ndarray
    :param device_tracking_array: Tracking data for camera (hand)
    :type device_tracking_array: List
    :param calibration_tracking_array: Tracking data for calibration target
    :type calibration_tracking_array: List
    :param left_rvecs: Vector of 3x1 ndarray, Rodrigues rotations for each
    camera
    :type left_rvecs: List[np.ndarray]
    :param left_tvecs: Vector of [3x1] ndarray, translations for each camera
    :type left_tvecs: List[np.ndarray]
    :param right_rvecs: Vector of 3x1 ndarray, Rodrigues rotations for each
    camera
    :type right_rvecs: List[np.ndarray]
    :param right_tvecs: Vector of [3x1] ndarray, translations for each camera
    :type right_tvecs: List[np.ndarray]
    :param quat_model2hand_array: Array of model to hand quaternions
    :type quat_model2hand_array: List
    :param trans_model2hand_array: Array of model to hand translaions
    :type trans_model2hand_array: List
    :return: Reprojection error, reconstruction error, left handeye matrix,
    left pattern to marker matrix, right handeye, right pattern to marker
    :rtype: float, float, np.ndarray, np.ndarray, np.ndarray, np.ndarray
    """
    # Do calibration
    left_handeye_matrix, left_pattern2marker_matrix =  \
        he.handeye_calibration(left_rvecs, left_tvecs, quat_model2hand_array,
                               trans_model2hand_array)

    right_handeye_matrix, right_pattern2marker_matrix =  \
        he.handeye_calibration(right_rvecs, right_tvecs, quat_model2hand_array,
                               trans_model2hand_array)

    # Filter common image points
    minimum_points = 10
    _, common_object_pts, common_l_image_pts, common_r_image_pts = \
    vu.filter_common_points_all_images(
        left_ids, left_object_points, left_image_points,
        right_ids, right_image_points,
        minimum_points)

    sse, num_samples = vm.compute_stereo_2d_err_handeye(
        common_object_pts,
        common_l_image_pts,
        left_camera_matrix, left_camera_distortion,
        common_r_image_pts,
        right_camera_matrix, right_camera_distortion,
        device_tracking_array, calibration_tracking_array,
        left_handeye_matrix, left_pattern2marker_matrix,
        right_handeye_matrix, right_pattern2marker_matrix
    )

    mse = sse / num_samples
    reproj_err = np.sqrt(mse)

    sse, num_samples = vm.compute_stereo_3d_err_handeye(
        l2r_rmat, l2r_tvec,
        common_object_pts,
        common_l_image_pts,
        left_camera_matrix, left_camera_distortion,
        common_r_image_pts,
        right_camera_matrix, right_camera_distortion,
        device_tracking_array, calibration_tracking_array,
        left_handeye_matrix, left_pattern2marker_matrix,
    )
    mse = sse / num_samples
    recon_err = np.sqrt(mse)

    return reproj_err, recon_err, \
        left_handeye_matrix, left_pattern2marker_matrix, \
        right_handeye_matrix, right_pattern2marker_matrix



# coding=utf-8

""" Command line processing for video calibration app. """

import argparse
from sksurgerycalibration import __version__
from sksurgerycalibration.ui.video_calibration_app import run_video_calibration


def main(args=None):
    """Entry point for simple video calibration application. """

    parser = argparse.ArgumentParser(description='videocalibration')

    parser.add_argument("-c", "--config",
                        required=True,
                        type=str,
                        help="Configuration file containing the parameters "
                             "(see config/video_chessboard_conf.json "
                             "for example).")

    parser.add_argument("-s", "--save",
                        required=False,
                        type=str,
                        help="Directory to save to.")

    parser.add_argument("-p", "--prefix",
                        required=False,
                        default="calib",
                        type=str,
                        help="Filename prefix to save to.")

    version_string = __version__
    friendly_version_string = version_string if version_string else 'unknown'
    parser.add_argument(
        "-v", "--version",
        action='version',
        version='sksurgerycalibration version ' + friendly_version_string)

    args = parser.parse_args(args)

    run_video_calibration(args.config, args.save, args.prefix)
